#include "comm.h"
#include <linux/netfilter.h>
#include <libipq.h>

#define BUFSIZE 65535



/* MESSAGE TYPES */

inline int msg_fin(struct tcphdr *tcphdr){
	return tcphdr->fin;
}


int syn_sending(struct iphdr *iphdr, struct tcphdr *tcphdr, struct hostent *host, int *list, int nr_process){
int i, ret=0;
	if (tcphdr->syn == 0)
		return -1;
	if(host == NULL){
		printf("Null host\n");
		return -1;
	}

	if(nr_process == 0)
		return 0;
	i = 0;
	while(host->h_addr_list[i] != 0){
		struct in_addr addr;
		memcpy(&addr, host->h_addr_list[i], host->h_length);
		if(iphdr->saddr == addr.s_addr){
			ret = 0;
			struct socket_cache_item data;
			memset(&data, 0, sizeof(data));
			data.ptr = (void *)list;
			data.nr_process = nr_process;
			data.saddr = iphdr->saddr;
			data.daddr = iphdr->daddr;
			data.sport = tcphdr->source;
			data.dport = tcphdr->dest;
			return have_to_register(&data);
		}
		i++;
	}
	return ret;
}

static inline int rst_receiving(struct tcphdr *tcphdr){
	return (tcphdr->rst);
}

static inline int first_ack_receiving(struct iphdr *iphdr, struct tcphdr *tcphdr, struct conntrack_struct *conn, u_int16_t size){
	return ((iphdr->daddr == conn->in_addr) &&
	        (tcphdr->ack) &&
		(conn->state & (S01|SLF|OTH)) &&
		size == 0);
}


static void die(struct ipq_handle *h)
{
        ipq_perror("passer");
        ipq_destroy_handle (h);
        exit(1);
}

int filter(struct hostent *host, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int *processes, int *num)
{
        int status, ret, pk_count=0;
        unsigned char buf[BUFSIZE];
        struct ipq_handle *h;
	struct iphdr *iphdr;
	struct tcphdr *tcphdr;
	struct conntrack_struct *conn;

        h = ipq_create_handle(0, PF_INET);
        if (!h)
                die(h);

        status = ipq_set_mode(h, IPQ_COPY_PACKET, BUFSIZE);
        if (status < 0)
                die(h);

        do{
                status = ipq_read(h, buf, BUFSIZE, 1000);
                if (status < 0)
                        die(h);

		if ((status == 0) && (host->h_length == 0)){
			printf("Filtering stopped by user\n");
			ipq_destroy_handle(h);
			return 0;
		}

                switch (ipq_message_type(buf)) {
                        case NLMSG_ERROR:
//                                fprintf(stderr, "Received error message %s\n",             strerror(ipq_get_msgerr(buf)));
                                break;

                        case IPQM_PACKET: {
				char *data;
				u_int16_t size;
			struct socket_cache_item datas;
                                ipq_packet_msg_t *m = ipq_get_packet(buf);
				pk_count++;

				iphdr = (struct iphdr *)m->payload;
				tcphdr = (struct tcphdr *)(m->payload + iphdr->ihl*4);
				data = (char *)(m->payload + 4*(iphdr->ihl+tcphdr->doff));
				size = m->data_len - 4*(iphdr->ihl+tcphdr->doff);

				printf("\n----------------------------------------------------------------------\n");
				printf("Message %d (%u : %d) -> (%u :%d) UAPRSF %d%d%d%d%d%d seq %u, ack %u, %d\n",
					pk_count,
					iphdr->saddr,
					tcphdr->source,
					iphdr->daddr,
					tcphdr->dest,
					tcphdr->urg,
					tcphdr->ack,
					tcphdr->psh,
					tcphdr->rst,
					tcphdr->syn,
					tcphdr->fin,
					ntohl(tcphdr->seq),
					ntohl(tcphdr->ack_seq),
					size);
				conn = find_conn(iphdr, tcphdr, cntab);
				if(conn == NULL){
					ret = syn_sending(iphdr, tcphdr, host, processes, *num);
					if (ret > 0){
						printf("A connect request from one of our processes\n");
						ret = try_add_conn(iphdr, tcphdr, cntab);
						if(ret > 0)
							printf("%d connections in space\n", ret);
					}
					else printf("Not found\n");
				}
 				else {
					printf("Found a related connection\n");
					if (rst_receiving(tcphdr)){
						printf("Reset notice\n");
						try_remove(conn, cntab, msg_buff);
					}
					else if(first_ack_receiving(iphdr, tcphdr, conn, size)){
						ret = finish_add(conn, tcphdr);
						printf("A connection established\n");
					}
					else if (tcphdr->syn == 0){// normal transfer
						ret = add_msg((char *)m->payload, m->data_len, conn, msg_buff);
					}
				}
				status = ipq_set_verdict(h, m->packet_id, NF_ACCEPT, 0, NULL);
				if (status < 0)
					die(h);
				break;
			}
			default:
				break;
                }
        } while (1);

        ipq_destroy_handle(h);
        return 0;
}
