#define __COMLIB__

#include "../utils/samory.h"
#include <pthread.h>
#include <signal.h>
#include "capture.h"
#include "dialogue.h"

/*
  This function is used like this:
  At migration time, restart fills up in_ and out_ with the saved values, along with
  acks and seqs
  A fake connection is established between the restarting process and another in the same
   host, mig_ first contains the port and address of the socket in this process side. It helps
    replay the message before  normal behavior is allowed.
    After connection is established and messages are all delivered, in_ contains the address and
    port for use by the filter, and corresponds to mig_
*/

int rebuild_conntable(int cfd, struct conntrack_struct *cntab, struct msg_buff *msg_buff){
struct conntrack_struct *conn, *prev;
struct msg_buff *buff, *bprev;
struct comm_header header;
int  i, conn_size, msg_size;

	if( read(cfd, &header,sizeof(header))!=sizeof(header)){
		printf("Did not receive enough data\n");
		return 0;
	}
	if (header.signature != MOD_MAGIC) {
		printf("Not a recognized file\n");
		return 0;
	}
	printf("Header read, %d connections\n", header.nr_conn);
// build connections linked list
	conn = cntab;
	while(conn->next != NULL)
		conn = conn->next;

	conn_size = sizeof(struct conntrack_struct);
	for (i = 0; i < header.nr_conn; i++){
		conn->next = (struct conntrack_struct *)malloc(conn_size);
		if (conn->next == NULL){
			printf("Cannot allocate cache space\n");
			close(cfd);
			return 0;
		}

		conn = conn->next;

		if( read(cfd, conn, conn_size)!= conn_size){
			printf("Did not receive enough data for open files\n");
			close(cfd);
			return 0;
		}
		conn->queue = NULL;
		conn->next = NULL;
		conn->state = 0;
	}
// build global message queue
	buff = msg_buff;
	while(buff->global_next != NULL)
		buff = buff->global_next;

	msg_size = sizeof(struct msg_buff);
	for (i = 0; i < header.nr_msg; i++){
		buff->global_next = (struct msg_buff *)malloc(msg_size);
		if(buff->global_next == NULL){
			printf("Cannot allocate cache space\n");
			close(cfd);
			return 0;
		}

		buff = buff->global_next;

		if( read(cfd, buff, msg_size)!= msg_size){
			printf("Did not receive enough data for open files\n");
			close(cfd);
			return 0;
		}
		buff->packet = (char *)malloc(buff->size);
		if(buff->packet == NULL){
			free(buff);
			printf("Cannot allocate cache space\n");
			close(cfd);
			return 0;
		}
		if( read(cfd, buff->packet, buff->size)!= buff->size){
			printf("Did not receive enough data for open files\n");
			close(cfd);
			return 0;
		}
// build local links
		buff->conn = conn_of_rank((int)buff->conn, cntab);
		if(buff->conn->queue == NULL)
			buff->conn->queue = buff;
		else {
			bprev = buff->conn->queue;
			while(bprev->local_next != NULL)
				bprev = bprev->local_next;
			bprev->local_next = buff;
		}
		buff->local_next = NULL;
	}
	return 1;
}


int verify_kernel_params(struct open_file *open_file, int nr_files, struct conntrack_struct *cntab, struct conns *sks){
int i, j;
struct socket_file *sock;
struct conntrack_struct *conn;

	if (cntab->next == NULL) return 0;

	j = 0; // counts the number of established connections

   	for(i = 0; i<nr_files; i++){
		if(open_file[i].type == SOCK_FILE){
			sock = &open_file[i].u.sock;
			if(sock->state == TCP_ESTABLISHED){
				printf("Looking for (%u : %u) -> (%u : %u).....",
				       sock->saddr, sock->sport, sock->daddr, sock->dport);
				conn = cntab->next;
				while(conn != NULL){
					printf("\nThis connection (%u : %u) -> (%u : %u)........",
					conn->in_addr, conn->in_port, conn->out_addr, conn->out_port);
					if((conn->in_addr == sock->saddr) && (conn->in_port == sock->sport) &&
						(conn->out_addr == sock->daddr) && (conn->out_port == sock->dport)){

						sks[j].cn = conn;
						sks[j].old = open_file[i].fd;
						j++;
						printf("stored in entry %d\n", j);
						break;
					}
					conn = conn->next;
				}
				if(conn == NULL){
					printf("Error file %d\n", open_file[i].fd);
					return -1;
				}
			}
		}
	}
	return j;
}


int restore_sockets(int skfd, int paramsfd, struct conns *sks, struct conntrack_struct *cntab, int nr_socket){
	printf("In restore_sockets\n");
	return restart_sockets(skfd, paramsfd, sks, cntab, nr_socket);
}


void install_filter(struct hostent *host, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int *pid_list, int *nr_process){
	filter(host, cntab, msg_buff, pid_list, nr_process);
}

/*
  We receive a notice from a partner.
  In case of a propagating notice, informing that checkpoint has been done and that there is no need
  to keep related messages, we update the connection and deleted related buffers
  In case of a checkpoint notify, we are given the new address. We then look for usable ports and bound
  some new sockets on them. We send that information to the partner. Later on we will listen to those sockets
*/

int handle_notify(int skfd, struct msg_struct *msg, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_process, int *pid_list){
struct conntrack_struct *conn;

	if (msg->type == CHK_MSG){
		printf("Received a checkpoint notice\n");
		if((conn = search(msg, cntab)) != NULL){
			printf("Found related connection\n");
			conn->yet_read = msg->prop;
		}
	}
	else if(msg->type == RST_MSG){
		printf("Going to restart notify handler\n");
		handle_restart(skfd, msg, cntab, msg_buff, nr_process, pid_list);
	}
	return 0;
}

int handle_restart_notify(int paramsfd, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_process, void *pid_list){
struct sockaddr_in	raddr;
struct msg_struct	msg;
int size;

	recvfrom(paramsfd, &msg, sizeof(msg), 0, (struct sockaddr *)&raddr, &size);
	printf("Received params\n");
	handle_notify(paramsfd, &msg, cntab, msg_buff, nr_process, pid_list);
}

int store_coms(struct conntrack_struct *cntab, struct msg_buff *msg_buff, struct open_file *open_files, int nr_files, int comfd, int skfd){
int ret;
	ret = check_connections(cntab, msg_buff, open_files, nr_files);
	if (ret == -1)
		return 0;

	ret = save_conn(cntab, msg_buff, comfd);
	if(ret != 0)
		return 0;

	notify_remote(CHK_MSG,skfd, cntab);
	return 1;
}

void replay_msg(struct conns *sks, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_sockets){
	replay_messages(sks, cntab, msg_buff, nr_sockets);
}

void zap_buffers(struct conns *sks, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_sockets){
	int i;
	for(i = 0; i < nr_sockets; i++){
		delete_message_queue(sks[i].cn->queue, msg_buff);
		sks[i].cn->queue = NULL;
	}
}
