#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>

// connection state
#define S01	1  // has issued a syn
#define EST	2  // is connected
#define	SLF	4  // this side has migrated
#define OTH	8  // remote end is migrated
#define F01	16 // has sent a fin
#define F02	32 // has received a fin

#define IN_PKT	1 	// incoming packets
#define OUT_PKT 2	// outgoing packets routing ok, I know the address

#define CHK_MSG 1	// checkpoint notice
#define RST_MSG 2	// restart notice
#define LST_MSG 3	// listening notice
// commands (message type between instances)
#define ADD	4	// adding a process to our list
#define REM	5	// removing a process
#define CKPT	6	// checkpoint request (before the normal time out)
#define RST	7	// restart request
#define QUIT	8	// uninstall the system
#define MIGR	9	// migration request

#define MAX_SOCKETS 8

#define FILTER_PORT	6503
#define NEW_PARAMS_PORT 6505


struct msg_buff;
/*
   This structure contains the state of a (about to) connected socket
*/
struct conntrack_struct {
	u_int32_t	in_addr; 	// were we are now
	u_int16_t	in_port;
	u_int32_t	out_addr; 	// where we think he is now
	u_int16_t	out_port;

// All sequence numbers are stored here in network byte order
	u_int32_t	yet_read; // Outgoing seqs read by remote aplication, remove from buffers at next checkpoint

	u_int32_t	copied_seq; // last byte read by the application,
	u_int32_t	snd_nxt; // next byte tcp was prepared to send, where we shall begin replay
	u_int32_t	rcv_nxt;
	u_int32_t	snd_una;

	struct msg_buff	*queue;
	struct conntrack_struct *next;
	short state; 			// state of the connection; see up
};

/*
	Structure used to pass messages between the filters
*/

struct msg_struct{
	int 		type;		// CHK_MSG or RST_MSG
	u_int32_t	in_addr; 	// were we are now
	u_int16_t	in_port;
	u_int32_t	mig_addr; 	// where we come from (for history)
	u_int16_t	mig_port;    // used to update the link after migration
	u_int32_t	out_addr;
	u_int16_t	out_port;
	u_int32_t 	prop;		// propagation pointer, or copied_seq
	u_int32_t	snd_nxt;
};

/*
  We store the messages traversing the net in a linked list per connection, and in a global list for serialization purposes. We store the message as given to us by the filter, but we only store out going
*/
struct msg_buff{
	struct msg_buff *global_next;// next message in global order
	struct msg_buff *local_next; // previous message
	struct conntrack_struct *conn; // connection back pointer
	u_int16_t size; // size of the hole ip packet
	char *packet; // the ip packet
};

struct conns{
int pid, fd, old, new;
/*u_int32_t saddr;
u_int16_t sport;
u_int32_t daddr;
u_int16_t dport;*/
u_int32_t yet_read;
u_int32_t copied_seq;
u_int32_t snd_nxt;
struct conntrack_struct *cn;
struct msg_buff *buffer;
void *file;
};

#ifdef __COMLIB__
/* Connections : retrieving */


int empty_list(struct conntrack_struct *conn){

	if (conn == NULL)
		return 1;
	if(conn->state != 0) // Something happened since creation
		return 0;
	return empty_list(conn->next);
}

/*
  This function returns not NULL if the message is to be tracked.
*/

struct conntrack_struct *find_conn(struct iphdr *iphdr, struct tcphdr *tcphdr, struct conntrack_struct *cntab){
struct conntrack_struct *conn;

	if(cntab->next == NULL){
		return NULL;
	}
	conn = cntab->next;
	while(conn != NULL){ // outgoing messages always have (s = in) and (d = out)
		conn->in_addr, conn->in_port, conn->out_addr, conn->out_port;

		if((((iphdr->saddr == conn->in_addr ) && (tcphdr->source == conn->in_port )) &&
		((iphdr->daddr == conn->out_addr) && (tcphdr->dest == conn->out_port))) ||
	        // incoming messages always have (s= out_addr) and (d = in)
		(((iphdr->saddr == conn->out_addr ) && (tcphdr->source == conn->out_port )) &&
		((iphdr->daddr == conn->in_addr) && (tcphdr->dest   == conn->in_port))))

			return conn;
		else
			conn = conn->next;
	}
		return NULL;
}


int packet_type(struct iphdr *iphdr, struct conntrack_struct *conn){

	if(iphdr->saddr != conn->in_addr )
		return IN_PKT;
	else
		return OUT_PKT;
}

/*
   We use these for modulo 32 bit arithmetics, typically tcp seq numbers
   Numbers are given in host byte order, and all results arge given
   in the same representation.
*/

inline int before(u_int32_t seq1, u_int32_t seq2){
        return (int)(seq1-seq2) < 0;
}

inline int after(u_int32_t seq1, u_int32_t seq2){
	return (int)(seq2-seq1) < 0;
}

/*
/* Notice: We receive a connecion header from abroad, and have to match it with one in our
/* data structure. Remind to match corresponding values, related to eventual migration
/*
/* when the link is up_to_date we have header (h.in = conn->out) .
/* else (when we receive a migration notice) h.in != h.mig.in, and h.mig points to conn->out.
/*    Here we know that we have to change out to h.in and adjust the state to OTH.
*/
struct conntrack_struct *search(struct msg_struct *header, struct conntrack_struct *cntab){
struct conntrack_struct *conn;

	if(cntab->next == NULL)
		return NULL;

	conn = cntab->next;

	printf("searching (%u : %u) -> (%u : %u)\n",
		header->out_addr, header->out_port, header->mig_addr, header->mig_port);

	while(conn != NULL){
		printf("This is (%u : %u) -> (%u : %u)\n",
			conn->in_addr, conn->in_port, conn->out_addr, conn->out_port);
		if((header->out_addr == conn->in_addr ) && (header->out_port == conn->in_port ) &&
		   (header->mig_addr == conn->out_addr) && (header->mig_port == conn->out_port))
			return conn;
		else
			conn = conn->next;
	}
    return NULL;
}

/* Connections: Creation and deletion */

/*
  Add a new connection tracking structure in the list, if necessary.
  It inspects packets that wish to establish a connection, to see if they are ours
*/

static int try_add_conn(struct iphdr *iphdr, struct tcphdr *tcphdr, struct conntrack_struct *cntab){
struct conntrack_struct *conn, *cn;
int nr_conn = 0;

	cn = cntab;
	while(cn->next != NULL){
		nr_conn++;
		cn = cn->next;
	}

	cn->next = (struct conntrack_struct *)malloc(sizeof(struct conntrack_struct));
	if(cn->next == NULL){
		printf("Cannot allocate space for %dth connection\n", nr_conn+1);
		return 0;
	}

	conn = cn->next;
	conn->state = S01;
	conn->in_addr = iphdr->saddr;
	conn->out_addr = iphdr->daddr;
	conn->in_port = tcphdr->source;
	conn->out_port = tcphdr->dest;

	conn->copied_seq = 0;
	conn->snd_una = tcphdr->seq;
	conn->snd_nxt = htonl(ntohl(tcphdr->seq) + 1);
	conn->rcv_nxt = 0;

	conn->queue = NULL;
	conn->next = NULL;
	return nr_conn;
}

/* We come here after receiving a SYN reply. So the connection is established */
static int finish_add(struct conntrack_struct *conn, struct tcphdr *tcphdr){
	if(conn->state & EST) // already established, duplicated message
		return 1;
	if(conn->state & (SLF|OTH)){
		conn->yet_read = tcphdr->ack_seq; // Normally it is ack_seq - 1
		conn->rcv_nxt = tcphdr->seq; // yes because a pure ack doesn't eat a seq number
		if(tcphdr->syn == 1)
			conn->rcv_nxt = htonl(ntohl(tcphdr->seq) +1);
		conn->state = EST;
		return 1;
	}
	if(conn->state != S01) //there is a problem !
		return 0;
	if(conn->snd_nxt != tcphdr->ack_seq){
		printf("Ack : %u does'nt match awaited : %u\n" ,ntohl(tcphdr->ack_seq), ntohl(conn->snd_nxt));
		return 0;
	}
	conn->state = EST;
	conn->yet_read = tcphdr->ack_seq; // Normally it is ack_seq - 1. Taken in account when cleaning the buffers
	conn->rcv_nxt = tcphdr->seq; // yes because a pure ack doesn't eat a seq number
	if(tcphdr->syn == 1)
		conn->rcv_nxt = htonl(ntohl(tcphdr->seq) +1);
	return 1;
}

int delete_message_queue(struct msg_buff *queue, struct msg_buff *msg_buff);

int try_remove(struct conntrack_struct *conn, struct conntrack_struct *cntab, struct msg_buff *msg_buff){
struct conntrack_struct *cn;

	cn = cntab;
	while (cn->next != conn)
		cn = cn->next;
	cn->next = conn->next;
	if(conn->queue != NULL)
		delete_message_queue(conn->queue, msg_buff);
	free(conn);
	return 1;
}

/* Message : Buffers and queues*/

static inline struct msg_buff* get_buff(int fd){
struct msg_buff *buff;

	buff = (struct msg_buff *)malloc(sizeof(struct msg_buff));
	if (buff == NULL)
		return NULL;
	if(read(fd, buff, sizeof(*buff)<sizeof(*buff))){
		free(buff);
		return NULL;
	}

	buff->packet = (char *)malloc(buff->size);
	if(buff->packet == NULL){
		free(buff);
		return NULL;
	}
	if(read(fd, buff->packet, buff->size) != buff->size){
		free(buff);
		return NULL;
	}
	buff->local_next = buff->global_next = NULL;
	return buff;
}

int add_msg(char *msg, u_int16_t global_size, struct conntrack_struct *conn, struct msg_buff *msg_buff){

	struct msg_buff *buff, *g_where, *l_where;
	struct tcphdr *tcphdr1, *tcphdr;
	struct iphdr *iphdr1, *iphdr;
	u_int16_t size;

	iphdr = (struct iphdr*)msg;
	tcphdr = (struct tcphdr *)(msg + iphdr->ihl*4);

	size = global_size - 4*(iphdr->ihl+tcphdr->doff);

	// We do not store incoming packets, instead we register their ending seq
	// We do not take in account messages arriving in wrong order.
	// Due to space problem, reordering is done when we see a message leaving this connection, that validates
	// an out-of-order message.
	if(packet_type(iphdr, conn) == IN_PKT){
		printf("Incoming message ...");
		if (conn->rcv_nxt == tcphdr->seq){
			printf("valid, stored\n");
		}
		else {
			printf("not valid, rcv_nxt is %u: and this was %u\n",
			ntohl(conn->rcv_nxt), ntohl(tcphdr->seq));
		}
		return 1;
	}

	// Outgoing packets
	// If the ack in this packet is greater than what we expected to receive, we have out-of-order packets
	// So adjust our variables!
	printf("Outgoing message\n");
	if (before(ntohl(conn->rcv_nxt), ntohl(tcphdr->ack_seq))){
		printf("Updating ack, now %u, \n", ntohl(tcphdr->ack_seq));
		conn->rcv_nxt = tcphdr->ack_seq;
	}

	if(size == 0){
		printf("Pure ack sending : %u, do not store \n", ntohl(tcphdr->ack_seq));
		return 1;
	}

	l_where = conn->queue;// l_where must point to the buffer just after this
	g_where = NULL; // g_where must point to the buffer just before this
	while(l_where != NULL){
		iphdr1 = (struct iphdr *)l_where->packet;
		tcphdr1 = (struct tcphdr *)(l_where->packet + iphdr1->ihl*4);
		printf("This buffer: seq = %u, ack = %u\n",ntohl(tcphdr1->seq), ntohl(tcphdr1->ack_seq));
		if(tcphdr->seq == tcphdr1->seq){
			printf("Duplicated message,(%u, %u) dropping\n",
			ntohl(tcphdr->seq), ntohl(tcphdr1->seq));
			return 1;
		}
		if(after(ntohl(tcphdr->seq), ntohl(tcphdr1->seq))){
			g_where = l_where;
			l_where = l_where->local_next;
		}
		else{
			printf("Alert, Out going message out-of-order\n");
			break;
		}
	}

	buff = (struct msg_buff *)malloc(sizeof(struct msg_buff));
	if(buff == NULL){
		printf("Can not allocate messages storage buffer\n");
		return 0;
	}

	buff->size = global_size;
	buff->global_next = buff->local_next = NULL;
	buff->conn = conn;
	buff->packet = (char *)malloc(global_size);
	if(buff->packet == NULL){
		printf("Can not allocate messages storage buffer\n");
		return 0;
	}

	memcpy(buff->packet, msg, global_size);
		printf("Global size was %d\n", global_size);
		if(strncmp(msg, buff->packet, global_size)!=0)
			printf("Two messages are different\n");

		iphdr = (struct iphdr *)(buff->packet);
		tcphdr = (struct tcphdr *)(buff->packet + iphdr->ihl*4);
		printf("Newly created buffer: seq = %u, ack = %u\n",ntohl(tcphdr->seq), ntohl(tcphdr->ack_seq));


	if((g_where == NULL) &&(l_where == NULL)){
		conn->queue = buff;
		g_where = msg_buff;
		while(g_where->global_next != NULL)
			g_where = g_where->global_next;
		g_where->global_next = buff;
		return 1;
	}

	else if((g_where != NULL) && (l_where !=NULL)){
		g_where->local_next = buff;
		buff->local_next = buff->global_next = l_where;
		g_where = msg_buff;
		while(g_where->global_next != l_where)
			g_where = g_where->global_next;
		g_where->global_next = buff;
		return 1;
	}

	else if ((g_where == NULL) && (l_where != NULL)){
		conn->queue = buff;
		buff->local_next = buff->global_next = l_where;
		g_where = msg_buff;
		while(g_where->global_next != l_where)
			g_where = g_where->global_next;
		g_where->global_next = buff;
		return 1;
	}

	else{ //(g_where != NULL && l_where == NULL)
		g_where->local_next = buff;
		while(g_where->global_next != NULL)
			g_where = g_where->global_next;
		g_where->global_next = buff;
		return 1;
	}
}

int delete_message_queue(struct msg_buff *queue, struct msg_buff *msg_buff){
// We must delete all the messages in a connection queue
// This occurs when the socket has sent and received a fin flag
// and after the applications have eaten all the data
// in our side copied_seq == last_seq and remote propgtd == out_mig_ack
// we follow the local buffers and fix the global pointers
	struct msg_buff *next_to_free, *to_free, *buff;

	if((msg_buff->global_next == NULL)||(queue == NULL)) return 1;

	buff = msg_buff;
	while(buff->global_next != queue)
		buff = buff->global_next;
	to_free = queue;
	while(to_free != NULL){
		next_to_free = to_free->local_next;
		buff->global_next = to_free->global_next;
		free(to_free);
		to_free = next_to_free;
		while(buff->global_next != to_free)
			buff = buff->global_next;
	}
	return 1;
}

/* packing and unpacking*/

static inline struct conntrack_struct *conn_of_rank(int rank, struct conntrack_struct *cntab){
int i;
struct conntrack_struct *conn;
	conn = cntab->next;
	if(rank <= 0)
		return NULL;
	i = 1;
	while((i < rank) && (conn != NULL)){
		conn = conn->next;
		i++;
	}
		return conn;
}

static inline int rank(struct conntrack_struct *cntab, struct conntrack_struct *conn){
struct conntrack_struct *cn;
int i;
	cn = cntab->next;
	i = 1;
	while((cn != NULL) && (cn != conn)){
	cn = cn->next;
	i++;
	}
	if(cn == NULL)
		return 0;
	return i;
}


int obsolete(struct msg_buff *buff){
struct iphdr *iphdr;
struct tcphdr *tcphdr;
u_int32_t last_seq;

	iphdr = (struct iphdr *)(buff->packet);
	tcphdr = (struct tcphdr *)(buff->packet + iphdr->ihl*4);
	last_seq = ntohl(tcphdr->seq) + buff->size - 4*(iphdr->ihl+tcphdr->doff);

	if(before(last_seq, ntohl(buff->conn->yet_read)))
		return 1;
	else
		return 0;
}

int out_propagated(struct msg_buff *buff){
struct iphdr *iphdr;
struct tcphdr *tcphdr;
u_int32_t last_seq;

	iphdr = (struct iphdr *)(buff->packet);
	tcphdr = (struct tcphdr *)(buff->packet + iphdr->ihl*4);
	last_seq = ntohl(tcphdr->seq) + buff->size - 4*(iphdr->ihl + tcphdr->doff);
	if(before(last_seq, ntohl(buff->conn->yet_read)))
		return 1;
	else
		return 0;
}

#endif
