
struct comm_header{
	int nr_conn;
	int nr_msg;
	unsigned long signature;
};

static int check_connections(struct conntrack_struct *cntab, struct msg_buff * msg_buff, struct open_file *open_files, int nr_files){
/*
after a checkpoint, this function saves the messages in the filter
Verifies that the sockets seen by checkpoint correspond to the connections seen here
Destroys messages already eaten by the application
call propagate
*/
struct conntrack_struct *conn;
struct socket_file *socket;
struct msg_buff *buff, *prev;
int conn_count, files;

	conn_count = 0;

	for(files = 0; files <nr_files; files++) {
		if(open_files[files].type == SOCK_FILE){
			socket = &(open_files[files].u.sock);
			if(socket->state == TCP_ESTABLISHED){
				printf("A socket file in established state\n");
				conn = cntab->next;
				while((conn->in_addr != socket->saddr) || (conn->in_port != socket->sport) ||
				     (conn->out_addr != socket->daddr) || (conn->out_port != socket->dport)){
					conn = conn->next;
					if (conn== NULL){
						printf("Couldn't find related connection!\n");
						return -1;
					}
				}

				conn_count++;
				printf("Found its related connection, %d\n", conn_count);

				conn->copied_seq = socket->copied_seq;
				conn->snd_nxt = socket->snd_nxt;
				conn->rcv_nxt = socket->rcv_nxt;
				conn->snd_una = socket->snd_una;
			}
		}
	}
	files = conn_count;
	conn_count = 0;
	//nr_msg = 0;
	conn = cntab->next;
	while(conn != NULL){
		if (conn->state == EST)
			conn_count++;
		conn = conn->next;
	}
	if(conn_count != files){
		printf("Error: Open sockets %d do not match communication table %d\n",files, conn_count);
		return -1;
	}
	// We now sweep buffers that are obsolete (those already read by remote aplication.
	// These have seq < yet_read

	prev = msg_buff;
	buff = msg_buff->global_next;
	while(buff != NULL){
		if(obsolete(buff)){
			struct msg_buff *old = buff;
			buff->conn->queue = buff->local_next;
			buff = buff->global_next;
			free(old);
		}
		else{
			prev = buff;
			buff = buff->global_next;
		}
	}

	return conn_count;
}


int replay_messages(struct conns *sks, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_sockets){
struct msg_buff *queue, *buff;
int remain, max_socket, ret, sent, r, i;
struct iphdr *iphdr;
struct tcphdr *tcphdr;
char *to_send, *rcv_buff;
int clilen, optval, optlen;
size_t size, window;
u_int32_t last_seq;
fd_set rfds;


// Now take on the replay process
// Idea: At re-connection time, the peer sent the value of his copied_seq (which I stored in yet_red), and snd_nxt
// So I have to begin replaying each buffers from sks[i].yet_read

// I also have my copied_seq which is the begining of what the last run did not eat.
// I have to save from her to snd_nxt in kernel buffers. The migrated will eventually eat them after restart

// To help replay and filter work, we have to isolate the buffers concerned by the replay process
// the filter will re-build them during the replay.

	printf("Unlinking buffers, %d sockets\n", nr_sockets);
	for(i = 0; i < nr_sockets; i++){
		sks[i].buffer = sks[i].cn->queue;
		sks[i].cn->queue = NULL;
// Now completely remove these buffers from the global list

		buff = msg_buff;
		queue = sks[i].buffer;

		printf("Unlinking socket n %d \n", i);

		if((msg_buff->global_next == NULL)||(queue == NULL)){
			printf("Nothing to do\n");
			break;
		}

		while(buff->global_next != queue){
			printf("searching head buffer\n");
			buff = buff->global_next;
		}
		printf("Got it\n");
		while(queue != NULL){
			printf("Got a buffer on queue\n");

			if(queue->global_next == NULL){
				printf("End of all buffers\n");
				buff->global_next = NULL;
				buff->local_next = NULL;
			}
			else if(queue->local_next == NULL){
				printf("End of this series\n");
				if(buff->global_next == queue)
					buff->global_next = queue->global_next;
			}
			else if(queue->global_next == queue->local_next){
				printf("Two buffers in the same list\n");
				buff->global_next = queue->global_next;
			}
			else{//
				printf("Two queues encountered\n");
				buff->global_next = queue->global_next;
				buff = buff->global_next;
			}
			queue  = queue->local_next;
		}
	}

// Make sure we have enough space to store all unread data in kernel
	for(i = 0; i < nr_sockets; i++){
		getsockopt(sks[i].new, SOL_SOCKET, SO_RCVBUF, &optval, &optlen);
		if(optval < (sks[i].snd_nxt - sks[i].copied_seq)){
			optval = ntohl(sks[i].snd_nxt) - ntohl(sks[i].copied_seq);
			optlen = sizeof(optval);
			setsockopt(sks[i].new, SOL_SOCKET, SO_RCVBUF, &optval, optlen);
		}
		printf("Data to store in socket %d is %d sized\n", i, optval);
	}

	printf("Replaying messages\n");
	remain = 0;
	max_socket = sks[nr_sockets -1].new + 1;

// First we set the sockets concerned by the replay process
// "En passant" we clean all unneeded buffers
	FD_ZERO(&rfds);

	for(i = 0; i < nr_sockets; i++){
		printf("Inspecting socket n %d buffers, yet_read is %u\n", i, ntohl(sks[i].yet_read));
		buff = sks[i].buffer;
		while(buff != NULL){ // There are backup-ed messages
			iphdr= (struct iphdr *)(buff->packet);
			tcphdr = (struct tcphdr *)(buff->packet + iphdr->ihl*4);
			size = buff->size - 4*(tcphdr->doff + iphdr->ihl);
			last_seq = ntohl(tcphdr->seq) + size;

			printf("begin : %u, size : %u, last_seq : %u\n", ntohl(tcphdr->seq), size, last_seq);

			if((after(ntohl(sks[i].yet_read), last_seq)) || (ntohl(sks[i].yet_read) == last_seq)){
				printf("This buffer was already eaten, flush\n");
				queue = buff;
				buff = buff->local_next;
				free(queue->packet);
				free(queue);
			}
			else{
				printf("Replay must start here on this connection\n");
				sks[i].buffer = buff;
				break;
			}
		}
		if(buff != NULL){
			printf("This socket %d is to be replayed\n", i);
			remain++;
		}
	}


		// send when ever possible. It works as follows
		// in each while loop, we look if it's better to receive before sending.
		// we know it by inspecting the ack_seq in the buffer, compared with snd_nxt
		// if ack_seq is greater we must first receive ack_seq - snd_nxt
		// if not we must from yet_read

	printf("Now starting replaying for %d sockets\n", remain);
	while(remain > 0){
	// Arriving here we know that we have something to send
		printf("%d queues left\n", remain);
		for(i = 0; i< nr_sockets; i++){

			printf("Inspecting socket n %d buffers\n", i);
			buff = sks[i].buffer;

			while(buff != NULL){ // we are still interested by this buffer
				iphdr= (struct iphdr *)(buff->packet);
				tcphdr = (struct tcphdr *)(buff->packet + iphdr->ihl*4);
				size = buff->size - 4*(tcphdr->doff + iphdr->ihl);
				last_seq = ntohl(tcphdr->seq) + size;

				printf("yet_read : %u, snd_nxt : %u, ack_seq : %u\n",
					ntohl(sks[i].yet_read), ntohl(sks[i].snd_nxt), ntohl(tcphdr->ack_seq));
				printf("begin : %u, size : %u, last_seq : %u\n", ntohl(tcphdr->seq), size, last_seq);

				if( after(last_seq, ntohl(sks[i].yet_read)) &&
				    before(ntohl(sks[i].snd_nxt), ntohl(tcphdr->ack_seq))){
				    	printf("Have to send this buffer\n");
					window = ntohl(sks[i].yet_read) - ntohl(tcphdr->seq);
					to_send = buff->packet + 4*(tcphdr->doff + iphdr->ihl) + window;
					size = buff->size - 4*(tcphdr->doff +iphdr->ihl) - window;

					sent = 0;
					while(sent < size){
						r= send(sks[i].new, to_send + sent, size - sent, 0);
						if(r == -1){
							fprintf(stderr, "send error in replay %s\n", strerror(errno));
							return -1;
						}
						sent += r;
					}

					sks[i].yet_read = htonl(ntohl(sks[i].yet_read) + size);
				// free this buffer
					sks[i].buffer = buff->local_next;
					free(buff->packet);
					free(buff);
				}
				else{
					printf("Have to read first\n");
					break;
				}
			}
		}
		printf("Prepare sockets to read\n");
		remain = 0;
		max_socket = 0;
		for(i = 0; i < nr_sockets; i++){
			if(before(ntohl(sks[i].snd_nxt), ntohl(sks[i].copied_seq))){
				printf("snd_nxt %u, copied %u : socket n %d still have to read\n",
					ntohl(sks[i].snd_nxt), ntohl(sks[i].copied_seq), i);
				FD_SET(sks[i].new, &rfds);
				remain++;
				if(max_socket < sks[i].new)
					max_socket = sks[i].new;
			}
		}
		// first prepare for read
		if(max_socket != 0){
			max_socket++;
			printf("In select\n");
			select(max_socket, &rfds, NULL, NULL, NULL);
		}

		printf("Now reading them\n");
		for(i = 0; i< nr_sockets; i++)
			if(FD_ISSET(sks[i].new, &rfds)){
				int rcv;
				rcv = ntohl(sks[i].copied_seq) - ntohl(sks[i].snd_nxt);
				rcv_buff = (char*)malloc(rcv);
				if((rcv = recv(sks[i].new, rcv_buff, rcv, 0)) < 0){
					fprintf(stderr, "receive error in replay %s\n", strerror(errno));
					return -1;
				}
				sks[i].snd_nxt = htonl(ntohl(sks[i].snd_nxt) + rcv);
				free(rcv_buff);
			}
		// update remain and fdset
		remain = max_socket = 0;
		for(i = 0; i< nr_sockets; i++){
			if((sks[i].buffer != NULL) || before(ntohl(sks[i].snd_nxt), ntohl(sks[i].copied_seq)))
				remain++;
		}
	}
	printf("Replay done, giving the new sockets to process\n");
	return 1;
}


// What to do if I migrate?
int restart_sockets(int skfd, int paramsfd, struct conns *sks, struct conntrack_struct *cntab, int nr_sk){
struct conntrack_struct *cn, *conn;
int i, nr_sockets, sk_index, sk_count, len, ret;
u_int32_t		addr;
struct sockaddr_in	caddr;
struct sockaddr_in	saddr;
struct msg_struct 	header;
struct samory_command	command;

	if(cntab->next == NULL)
		return 1;

	cn = cntab->next;
	sk_count = 0;
	printf("in restart_sockets\n");
	while(cn != NULL){

	// first register my new address
		len = sizeof(caddr);
		getsockname(skfd, (struct sockaddr*)&caddr, &len);

// Build a header for a migration notice (port is 0 because it may change)
		memset(&header, 0, sizeof(header));
		header.out_addr = cn->out_addr;  // his addr
		header.out_port = 0;
		header.mig_addr = cn->in_addr;   // my old adddr
		header.mig_port = 0;
		header.in_addr = caddr.sin_addr.s_addr; // my new addr
		header.in_port = 0;

//		sks[sk_count].cn = cn;
		sk_count++;

// Count the connections concerned by these source and destination
		conn = cn->next;
		nr_sockets = 1;
		printf("Found a connection to migrate, .....");

		while(conn != NULL){
			if((conn->out_addr == cn->out_addr) &&(conn->in_addr == cn->in_addr)){
				nr_sockets++;
//				sks[sk_count].cn = conn;
				sk_count++;
				conn->in_addr = header.in_addr;
				conn->state = SLF;
			}
			conn = conn->next;
		}
		printf("%d sockets from %u to %u \n", nr_sockets, cn->in_addr, cn->out_addr);
		cn->in_addr = header.in_addr; // Do this at last because of the test up
		cn->state = SLF; // mark it as migrated

// Send the notice to peer
		saddr.sin_family = AF_INET;
		saddr.sin_addr.s_addr = cn->out_addr;
		saddr.sin_port = htons(FILTER_PORT);

		header.type = RST_MSG;
		command.type = RST_MSG;
		printf("sending restart notice .............");
		ret = sizeof(command);
		if(sendto(skfd, &command, ret, 0, (struct sockaddr *)&saddr, sizeof(saddr)) != ret){
				fprintf(stderr, "send error %s\n", strerror(errno));
				return -1;
			}
		printf("Done\n");

		saddr.sin_port = htons(NEW_PARAMS_PORT);
		printf("Now sending parameters .............");
		ret = sizeof(header);
		if(sendto(paramsfd, &header, ret, 0, (struct sockaddr *)&saddr, sizeof(saddr)) != ret){
			fprintf(stderr, "send error %s\n", strerror(errno));
			return -1;
		}
		printf("Done \n");

// Wait the new port for each connection and send a connection request to it
		for(i = 0; i < nr_sockets; i++){
			printf("Waiting for new parameters .........");
			ret = recvfrom(paramsfd, &header, sizeof(header), 0, (struct sockaddr *)&saddr, &len);
			if (ret != sizeof(header)){
				printf("sorry, did not receive enough data, leaving\n");
				return -1;
			}
			if(header.type != LST_MSG){
				printf("Protocol error, leaving\n");
				return -1;
			}
// here have received some new parameters(my_new_addr, his_old_addr, his_new_addr = in, mig, out)
			printf("Ok : (%u : %u) -> (%u : %u)\n",
				header.out_addr, header.out_port, header.in_addr, header.in_port);

// Find the connection concerned by this
			conn = cn; // always begin here because upper part is supposed to be done
			printf("Searching (%u : %u) -> (%u : %u)",
				header.in_addr, header.in_port, header.mig_addr, header.mig_port);
			while(conn != NULL){
				printf("\nThis is (%u : %u) -> (%u : %u)",
					conn->in_addr, conn->in_port, conn->out_addr, conn->out_port);
				if((header.in_addr == conn->in_addr ) && (header.in_port == conn->in_port) &&
	    			   (header.mig_addr == conn->out_addr) && (header.mig_port == conn->out_port))
						break;
					else
						conn = conn->next;
			}
			if(conn == NULL){
				printf("Cannot find related connection\n");
				return -1;
			}

			printf("..........Ok\nSearching entry in sockets table, (%d elements)....", nr_sk);

			sk_index = 0;
			while (sk_index < nr_sk){
				printf("\nsocket file : entry %d, descriptor %d.....", sk_index, sks[sk_index].old);
				if(sks[sk_index].cn == conn)
					break;
				sk_index++;
			}

			if(sk_index == nr_sk){
				printf("Not found\n");
				return -1;
			}

			printf("Ok, entry %d.\n", sk_index);
			// remind these values, for replay

			sks[sk_index].yet_read = header.prop;
			sks[sk_index].snd_nxt = header.snd_nxt;
			sks[sk_index].copied_seq = conn->copied_seq;


			caddr.sin_family = AF_INET;
			caddr.sin_addr.s_addr = header.in_addr;
			caddr.sin_port = conn->in_port;

			if((sks[sk_index].fd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
				printf("Can't open stream socket\n");
				return -1;
			}
			if(bind(sks[sk_index].fd, (struct sockaddr*)&caddr, sizeof(caddr))<0){
				fprintf(stderr, "Can't bind local address choose any\n");
				caddr.sin_port = INADDR_ANY;
				bind(sks[sk_index].fd, (struct sockaddr*)&caddr, sizeof(caddr));
			}

			len  = sizeof(caddr);
			getsockname(sks[sk_index].fd, (struct sockaddr*)&caddr, &len);

			printf("Now sending replay parameters .........");

			header.prop = conn->copied_seq; // Yes this has been set at last checkpoint,
			header.snd_nxt = conn->snd_nxt; // which is the one we are using
			header.in_addr = caddr.sin_addr.s_addr;
			header.in_port = caddr.sin_port;

			ret = sizeof(header);
			if(sendto(paramsfd, &header, ret, 0, (struct sockaddr *)&saddr, sizeof(saddr)) != ret){
				fprintf(stderr, "send error %s\n", strerror(errno));
				return -1;
			}
			printf("Done \n");

// Now prepare the conn_struct for reopen files.
			conn->out_addr = header.out_addr;
			conn->out_port = header.out_port;
			conn->in_addr = caddr.sin_addr.s_addr;
			conn->in_port = caddr.sin_port;
/*			sks[sk_index].sport = conn->in_port;
			sks[sk_index].dport = conn->out_port;
			sks[sk_index].saddr = conn->in_addr;
			sks[sk_index].daddr = conn->out_addr;
*/
// prepare an address structure for a connection
			saddr.sin_family = AF_INET;
			saddr.sin_addr.s_addr = header.out_addr;
			saddr.sin_port = header.out_port;

			printf("Now connecting (%u : %u)->(%u : %u)\n",
				caddr.sin_addr.s_addr, caddr.sin_port, saddr.sin_addr.s_addr, saddr.sin_port);
			if(connect(sks[sk_index].fd,(struct sockaddr *)&saddr, sizeof(saddr)) == -1){
				fprintf(stderr, "connect error %s\n", strerror(errno));
				return -1;
			}

		}

// Go to the next group of connections
		while(cn != NULL){
			if(cn->state == 0) // this means not yet seen
				break;
			cn = cn->next;
		}
	}
	if(sk_count != nr_sk)
		printf("Warning sockets numbers do not match !\n");
	return sk_count;
}

/*
  Restarting sockets:
  First, the migrating part sends a notice and wait for the port to send a connection request to,
  for each socket connecting these two sites.
  The receiver then bounds the specified sockets to any free port and issue a listen on it
  It sends that information and the function returns
  somelater time, it gets a message notifying that restart is beginning truly.
  It then waits for a client connection on all the listening sockets
  When a connection is established, it closes the first socket and returns.
*/
int handle_restart(int skfd, struct msg_struct *header, struct conntrack_struct *cntab, struct msg_buff *msg_buff, int nr_process, int *pid_list){

struct conntrack_struct *conn;
int i, ret, clilen;
struct sockaddr_in	caddr;
struct sockaddr_in	saddr;
struct sockaddr_in	name;
socklen_t 		namelen;

struct conns sks[MAX_SOCKETS];
struct socket_cache_item sk_cache;
struct msg_struct msghdr;

struct timeval wait_time;

	i = 0;
	conn = cntab->next;

	while(conn != NULL){
// First track the connections concerned by this migration
		if((conn->out_addr == header->mig_addr) && (conn->in_addr == header->out_addr)){
// Find a new port for restarting socket
			saddr.sin_family = AF_INET;
			saddr.sin_port = htons(0);
			saddr.sin_addr.s_addr = conn->in_addr;

			if((sks[i].old = socket(AF_INET, SOCK_STREAM, 0)) < 0){
				fprintf(stderr, "Can't open stream socket %s\n", strerror(errno));
				return -1;
			}

			if(bind(sks[i].old, (struct sockaddr*)&saddr, sizeof(saddr)) < 0){
				fprintf(stderr, "Can't bind local address %s\n", strerror(errno));
				return -1;
			}

			if(listen(sks[i].old, 1) == -1){
				fprintf(stderr, "Listen error %s\n", strerror(errno));
				return -1;
			}

// Freeze the original socket while doing modifications
			printf("Parameters build, now freeze the original socket\n");

			memset(&sk_cache, 0, sizeof(sk_cache));
			sk_cache.ptr = (void *)pid_list;
			sk_cache.nr_process = nr_process;
			sk_cache.saddr = conn->in_addr;
			sk_cache.sport = conn->in_port;
			sk_cache.daddr = conn->out_addr;
			sk_cache.dport = conn->out_port;

			if(freeze_socket(&sk_cache) >= 0 )
				printf("Done\n");
			else{
				printf("Socket not found\n");
				return -1;
			}
			//remind this freeze parameters, useful for exchange and unfreeze
			sks[i].pid = sk_cache.pid;
			sks[i].fd = sk_cache.fd;
			sks[i].file = sk_cache.ptr;
			sks[i].copied_seq = sk_cache.copied_seq;

// Collect and send the new socket params to peer

			namelen  = sizeof(name);
			if(getsockname(sks[i].old, (struct sockaddr*)&name, &namelen) == -1){
				fprintf(stderr, "Getsockname error %s\n", strerror(errno));
				unfreeze_socket(&sk_cache);
				return -1;
			}

			memset(&msghdr, 0, sizeof(msghdr));

		// We send these parameters(his_new_addr, my_old_addr, my_new_addr = in, mig, out)

			msghdr.type = LST_MSG;
			msghdr.in_addr = header->in_addr;
			msghdr.in_port = conn->out_port; // to help search at arrival
			msghdr.mig_addr = conn->in_addr;
			msghdr.mig_port = conn->in_port;
			msghdr.out_addr = name.sin_addr.s_addr;
			msghdr.out_port = name.sin_port;
			msghdr.prop = sk_cache.copied_seq;
			msghdr.snd_nxt = conn->snd_nxt;

			printf("Sending new parameters to pair : (%u : %u) -> (%u : %u)\n",
				msghdr.out_addr, msghdr.out_port, msghdr.in_addr, msghdr.in_port);

			ret = sizeof(msghdr);

			caddr.sin_family = AF_INET;
			caddr.sin_port = htons(NEW_PARAMS_PORT);
			caddr.sin_addr.s_addr = msghdr.in_addr;
			clilen = sizeof(caddr);

			if(sendto(skfd, &msghdr, ret, 0, (struct sockaddr *)&caddr, clilen) != ret){
				fprintf(stderr, "send error %s\n", strerror(errno));
				unfreeze_socket(&sk_cache);
				return -1;
			}

			printf("Waiting for replay parameters\n");
			if(recvfrom(skfd, &msghdr, ret, 0, (struct sockaddr *)&caddr, &clilen) != ret){
				fprintf(stderr, "recvfrom error %s\n", strerror(errno));
				unfreeze_socket(&sk_cache);
				return -1;
			}

			sks[i].yet_read = msghdr.prop;
			sks[i].snd_nxt = msghdr.snd_nxt;
// update our tables
			conn->out_addr = msghdr.in_addr;
			conn->out_port = msghdr.in_port;
			conn->in_addr = name.sin_addr.s_addr;
			conn->in_port = name.sin_port;
			conn->state = OTH;

			printf("Has read %u, and was sending %u\n",
				ntohl(msghdr.prop), ntohl(msghdr.snd_nxt));
// Now wait for peer to connect on newly created socket
			clilen = sizeof(caddr);
			printf("Waiting for a connection\n");
			sks[i].new = accept(sks[i].old,(struct sockaddr *)&caddr, &clilen);
			if(sks[i].new == -1){
				fprintf(stderr, "Accept error %s\n", strerror(errno));
				unfreeze_socket(&sk_cache);
				return -1;
			}
			printf("Client connected\n");

			sks[i].cn = conn;
// close the listening socket
			close(sks[i].old);
			sks[i].old = 0;// remove from our list
			i++;
		}
		conn = conn->next;
	}

	ret = i;
	replay_messages(sks, cntab, msg_buff, ret);
	printf("Replay done, giving the new sockets to process\n");

// Then give the sockets to the application
	for(i = 0; i < ret; i++){
		sk_cache.pid = sks[i].pid;
		sk_cache.fd = sks[i].fd;
		sk_cache.ptr = sks[i].file;
		sk_cache.cur_fd = sks[i].new;
		sk_cache.nr_process = 0;
		printf("Exchanging the sockets\n");
		xchange_sock(&sk_cache);
		printf("And unfreeing the frozen one\n");
		unfreeze_socket(&sk_cache);
	}

	printf("All went good, allowing normal execution\n");
	return 1;
}


/*
/* Notify_remote : Tell the remote filter that the migration or checkpoint is done,
/*                 We walk through the connections and send the new parameters to each of the filters
/*                 For migration: Send the new address and port, and propgtd
/*                 For propagation, only send the new propgtd, so and msg_struct is sufficient
/*                 TODO See how to group communications to the same destination in one send
*/

int notify_remote(int which, int skfd, struct conntrack_struct *cntab){
struct sockaddr_in skaddr;
char buff[MAX_MSG_SIZE];
struct msg_struct msg;
struct samory_command command;
struct conntrack_struct *conn;

	if(cntab->next == NULL)
		return 0;

	conn = cntab->next;
	skaddr.sin_family = AF_INET;
	skaddr.sin_port = htons(FILTER_PORT);

	while (conn){
		skaddr.sin_addr.s_addr = conn->out_addr;
		msg.mig_addr = conn->in_addr;
		msg.mig_port = conn->in_port;
		msg.out_addr = conn->out_addr;
		msg.out_port = conn->out_port;
		msg.prop = conn->copied_seq;
		msg.type = which;
		memset(&command, 0, sizeof(command));
//		command.type = which;
		memcpy(&command, &msg, sizeof(msg));
		printf("Notifying %u of checkpoint\n", conn->out_addr);
		sendto(skfd, &command, sizeof(command), 0, (struct sockaddr *)&skaddr, sizeof(skaddr));
		conn = conn->next;
	}
	return 0;
}


/*
These functions save the connections structure and all the messages as from the last checkpoint.
The calling routine has marked the connections to see which messages can be ignored (those already eaten by the process). The connections are stored in the file whose descriptor is given.
*/

int save_conn(struct conntrack_struct *cntab, struct msg_buff *msg_buff, int comm_fd){
struct msg_buff *buff, *prev;
struct conntrack_struct *conn;
struct comm_header header;
int nr_conn, nr_msg, ret;
/*
register the actual limit of propagated messages
sends that informations to groupmates
informs his counterpart (no need to wait for an ack)
*/

	printf("Cleaning useless buffers\n");
	buff = msg_buff->global_next;
	prev = msg_buff;
	while(buff!= NULL){
		if(obsolete(buff)){
			struct msg_buff *old = buff;
			printf("Deleting a buffer\n");
			prev->global_next = buff->global_next;
			if(prev->local_next == buff)
				prev->local_next = buff->local_next;
			if(buff->conn->queue == buff)
				buff->conn->queue = buff->local_next;

			buff = buff->global_next;

			free(old->packet);
			free(old);
		}
		else{
			prev = buff;
			buff = buff->global_next;
		}
	}

	printf("Now saving the connections\n");

	nr_conn = nr_msg = 0;
	conn = cntab->next;
	buff = msg_buff->global_next;

	while(conn != NULL){
		if(conn->state >= EST)
			nr_conn++;
		conn = conn->next;
	}
	while(buff != NULL){
		nr_msg++;
		buff = buff->global_next;
	}
	header.nr_conn = nr_conn;
	header.nr_msg = nr_msg;
	header.signature = MOD_MAGIC;

	ret = write(comm_fd, &header, sizeof(header)); // TODO replace by mcast primitive?
	if(ret != sizeof(header)){
		printf("Error writing in communications file, wrote %d\n", ret);
		return -1;
	}

	conn = cntab->next;
	while(conn != NULL){
		if(conn->state >= EST)
			ret = write(comm_fd, conn, sizeof(*conn));
			if(ret != sizeof(*conn)){
				printf("Error storing %d connection\n", rank(cntab, conn));
				return -1;
			}
		conn = conn->next;
	}

	buff = msg_buff->global_next;
	while(buff != NULL){
		conn = buff->conn;
		(int)buff->conn = rank(cntab, conn);
		ret = write(comm_fd, buff, sizeof(struct msg_buff));
		if(ret != sizeof(struct msg_buff)){
			printf("Error storing message header\n");
			return -1;
		}
		ret = write(comm_fd, buff->packet, buff->size);
		if(ret != buff->size){
			printf("Error storing message\n");
			return -1;
		}
		buff->conn = conn;
		buff = buff->global_next;
	}
return 0;
}
