int open_socket(struct socket_file *socks, int fd, int nr_sockets, struct conns *sks, struct hostent *host)
{
int	desc, optval, optlen, j;
struct sockaddr_in my_addr, r_addr;

	printf("Socket\n");
	if(socks->state != TCP_ESTABLISHED){

		j = 0;
		while(j<nr_sockets){
			if(sks[j].old == fd){
				printf("Descriptor conflict, %d\n", sks[j].old);
				desc = dup(sks[j].fd);
				if(desc == -1){
					fprintf(stderr, "Dup %s\n", strerror(errno));
					return -1;
				}
				close(sks[j].fd);
				sks[j].fd = desc;
				break;
			}
			j++;
		}
		// Here we know that the wanted descriptor is free
		my_addr.sin_family = socks->domain;
		memcpy(&my_addr.sin_addr, host->h_addr, host->h_length);
		my_addr.sin_port = socks->sport;
		if (socks->saddr == 0)	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		if (socks->sport == 0)	my_addr.sin_port = htons(0);

		r_addr.sin_family = socks->domain;
		r_addr.sin_addr.s_addr = socks->daddr;
		r_addr.sin_port = socks->dport;

		if ((desc = socket(socks->domain, socks->type, socks->protocol)) < 0)
		{
			perror("Can't reopen the socket\n");
			return -fd;
		}
		printf("Socket returned descriptor: %d, Wanted descriptor %d\n", desc, fd);

		if ((socks->saddr != 0)||(socks->sport !=0))
			if(bind(desc, (struct sockaddr*) &my_addr, sizeof(my_addr)) < 0){
				perror("Can't bind local address\n");
				return -fd;
			}

		if(socks->state == TCP_LISTEN){
			printf("In listen state\n");
			if(bind(desc,(struct sockaddr*)&my_addr, sizeof(my_addr))<0){
				perror("Can't bind local address\n");
				return -fd;
			}
			if (listen(desc, socks->backlog)<0){
				perror("Listen error\n");
				return -fd;
			}
		}
		if (desc!=fd){
			printf("Returned descriptor %d, force opening the socket with descriptor %d\n",desc, fd);
			if (dup2(desc, fd) != fd){
				perror("Could not apply given descriptor\n");
				return -fd;
			}
			close(desc);
		}
	}
	else{ //(socks->state == TCP_ESTABLISHED)
	// already opened, find it in sks
		j = 0;
		while((j<nr_sockets) && (sks[j].old != fd)){
			j++;
		}
		if(j == nr_sockets){
			printf("Could not find the socket !\n");
			return -fd;
		}
		if(sks[j].fd != fd){
			int k;
			printf("Descriptors do not match, shall dup2\n");
			// first look if this descriptor is in sks; in case move it; otherwise close it
			k = 0;
			while((k<nr_sockets) && (sks[k].fd != fd)){
				k++;
			}
			if(k == nr_sockets){
				printf("Not in sks, can close it\n");
				close(fd);
			}
			else{
				printf("Have to move it\n");
				desc = dup(sks[k].fd);
				if(desc == -1){
					fprintf(stderr, "Dup %s\n", strerror(errno));
					return -1;
				}
				close(sks[k].fd);
				sks[k].fd = desc;
			}
			if(dup2(sks[j].fd, fd) != fd){
				fprintf(stderr, "Dup2 %s\n", strerror(errno));
				return -fd;
			}
			close(sks[j].fd);
			sks[j].fd = fd;
		}
	}
	return fd;
}

int open_dup(struct dup_file *dups, int fd)
{

	if(fd<3){
		printf("file descriptor %d already in use\n", fd);
		return (fd);
	}
	printf("Duplicating %d with descriptor %d\n", dups->orig_fd, fd);
	if(dup2(dups->orig_fd, fd)!=fd){
		printf("Unable to reopen %d\n",fd);
		return -fd;
	}
	printf("Opening ok\n");
	return fd;
}

int open_pipe(struct pipe_file *pipes, int fd)
{
  printf("Pipe file, descriptor %d;   ", fd);
  printf("size %d;   ", pipes->i_size);
  printf("R/W %d\n", pipes->rdwr);
}

int open_reg(struct regular_file *reg_f, int fd)
{
int desc;
	if(fd<3){
		printf("file descriptor %d already in use\n", fd);
		return (fd);
	}
	printf("Opening regular file %s with descriptor %d\n", reg_f->filename, fd);
	if((desc = open(reg_f->filename, reg_f->flags, reg_f->mode))!=fd){
		if(dup2(desc, fd)!=fd){
			printf("Could not reopen the file\n");
			return (-fd);
		}
		close(desc);
		printf("File reopened wit dup2\n");
	}
	if(lseek(fd, reg_f->file_pos, SEEK_SET)==-1){
		printf("Warning : %d needs to be updated\n", fd);
		return(fd);
	}
	printf("Ok..............\n");
	return (fd);
}

int reopen_files(struct open_file *open_file, int nr_files, int nr_sockets, struct conns *sks, struct hostent *host){
int i, fd, j, desc;
// This file list is sorted; lowest descriptors are fisrt

   for(i = 0; i<nr_files; i++){
	j = 0;
	while(j<nr_sockets){
		if((sks[j].old == open_file[i].fd) && (open_file[i].type != SOCK_FILE)){
			printf("Descriptor conflict, %d\n", sks[j].old);
			fd = dup(sks[j].fd);
			if(fd == -1){
				fprintf(stderr, "Dup %s\n", strerror(errno));
				return -1;
			}
			close(sks[j].fd);
			sks[j].fd = fd;
			break;
		}
		else if (open_file[i].type == SOCK_FILE)
			break;
		j++;
	}
   	printf("reopening file descriptor %d\n", open_file[i].fd);

	switch(open_file[i].type){
	   case REG_FILE:
		if ((fd = open_reg(&open_file[i].u.file, open_file[i].fd) < 0)){
		   printf("Could not open file descriptor %d\n", -fd);
		   return -1;
		}
		break;
	   case DUP_FILE:
		if ((fd = open_dup(&open_file[i].u.dup, open_file[i].fd) < 0)){
		   printf("Could not open file descriptor %d\n", -fd);
		   return -1;
		}
		break;
	   case PIPE_FILE:
		if ((fd = open_pipe(&open_file[i].u.pipes, open_file[i].fd) < 0)){
		   printf("Could not open pipe descriptor %d\n", -fd);
		   return -1;
		}
		break;
	   case SOCK_FILE:
		if ((fd = open_socket(&open_file[i].u.sock, open_file[i].fd, nr_sockets, sks, host) < 0)){
		   printf("Could not open socket descriptor %d\n", -fd);
		   return -1;
		}
		break;
	}

	if (fcntl(open_file[i].fd, F_SETFD, 0)!= -1)
		printf("File %d successfully opened\n", open_file[i].fd);
   }
   return 0;
}
