#include <linux/smp.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/sys.h>
#include <linux/sched.h>
#include <asm/page.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <linux/mm.h>
#include <asm/mman.h>
#include <linux/slab.h>
#include <asm/types.h>
#include <linux/file.h>
#include <linux/ptrace.h>
#include <linux/time.h>

#include <linux/inet.h>
#include <linux/socket.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/net.h>
#include <net/sock.h>

#include <linux/skbuff.h>

#include "samory.h"
#include "config.h"

int store(struct file *f, void *data, int len);

/*
#define DO_NO_PAGE 0xc01f0510
#define DO_SWAP_PAGE 0xc01f02d0
*/

EXPORT_NO_SYMBOLS;
MODULE_LICENSE("GPL");

struct timeval start_time, end_time;

/**************** The checkpoint utility functions begin from here**************/

int (*do_no_page)(struct mm_struct * mm,
		struct vm_area_struct * vma,
		unsigned long address,
		int write_access,
		pte_t *page_table) = (void*)DO_NO_PAGE;

int (*do_swap_page)(struct mm_struct * mm,
		struct vm_area_struct * vma,
		unsigned long address,
		pte_t * page_table,
		swp_entry_t entry,
		int write_access) = (void*)DO_SWAP_PAGE;

int store(struct file *f, void *data, int len) {

int ret, wrtn = 0;
mm_segment_t fs;

	fs= get_fs();
	set_fs(KERNEL_DS);
	while (wrtn<len) {
		ret = f->f_op->write(f, data+wrtn, len-wrtn, &(f->f_pos));
		if(ret<=0){
			printk(KERN_DEBUG "transfer error\n");
			set_fs(fs);
			return ret;
		}
		else wrtn+= ret;
	}
	set_fs(fs);
	return wrtn;
}

static inline char * mem_prot(unsigned long prot, char *strings)
{
  memcpy(strings,"----\0", 5);
  if (prot & PROT_READ) strings[0] = 'R';
  if (prot & PROT_WRITE) strings[1] = 'W';
  if (prot & PROT_EXEC) strings[2] = 'X';
  if (prot & PROT_NONE) strings[3] = 'N';
  return strings;
}

static inline char * mem_flags(unsigned long flags, char *strings)
{
  memcpy(strings,"---- ---- ---- ---\0", 19);
  if (flags & VM_READ) strings[0] = 'r';
  if (flags & VM_WRITE) strings[1] = 'w';
  if (flags & VM_EXEC) strings[2] = 'x';
  if (flags & VM_SHARED) strings[3] = 's'; else strings[3] = 'p';
  if (flags & VM_MAYREAD) strings[5] = 'R';
  if (flags & VM_MAYWRITE) strings[6] = 'W';
  if (flags & VM_MAYEXEC) strings[7] = 'X';
  if (flags & VM_MAYSHARE) strings[8] = 'S';
  if (flags & VM_GROWSDOWN) strings[10] = 'd';
  if (flags & VM_GROWSUP) strings[11] = 'u';
  if (flags & VM_SHM) strings[12] = 's';
  if (!(flags & VM_DENYWRITE)) strings[13] = 'w';
  if (flags & VM_EXECUTABLE) strings[15] = 'X';
  if (flags & VM_LOCKED) strings[16] = 'L';
  if (flags & VM_IO) strings[17] = 'I';
  return strings;
}

static inline int get_header(struct file *f, struct task_struct *p, int check) {
	
	struct process_header header;
	int i;
  if (!check) return 0;

	printk(KERN_DEBUG "Processing the header\n");

	header.flags = p->flags;
	header.sigpending = p->sigpending;
	header.personality = p->personality;
	header.did_exec = p->did_exec;
	header.pid = p->pid;
	header.pgrp = p->pgrp;
	header.tty_old_pgrp = p->tty_old_pgrp;
	header.session = p->session;
	header.tgid = p->tgid;
	header.leader = p->leader;
	header.sas_ss_sp = p->sas_ss_sp;
	header.sas_ss_size = p->sas_ss_size;

	if (p->p_opptr==NULL)
	  header.family[0] = 0;
	else
	  header.family[0]= p->p_opptr->pid;
	if (p->p_pptr==NULL)
	  header.family[1] = 0;
	else
	  header.family[1] = p->p_pptr->pid;
	if (p->p_cptr==NULL)
	  header.family[2] = 0;
	else
	  header.family[2] = p->p_cptr->pid;
	if (p->p_ysptr==NULL)
	  header.family[3] = 0;
	else
	  header.family[3] = p->p_ysptr->pid;
	if (p->p_osptr==NULL)
	  header.family[4] = 0;
	else
	  header.family[4] = p->p_osptr->pid;
/* User and group infos */
	header.uid = p->uid;
	header.euid = p->euid;
	header.suid = p->suid;
	header.fsuid = p->fsuid;
	header.gid = p->gid;
	header.egid = p->egid;
	header.sgid = p->sgid;
	header.fsgid = p->fsgid;
	header.ngroups = p->ngroups;
	for (i=0; i<p->ngroups; i++)
	  header.groups[i] = p->groups[i];
	memcpy(header.comm,p->comm,str_len(p->comm));
	header.comm[str_len(p->comm)] = '\0';
	header.locks = p->locks;

		if (store(f, &header, sizeof(header))!= sizeof(header)){
		  printk(KERN_DEBUG "Problem storing the header\n");
		  return -1;
		}
		printk(KERN_DEBUG "Header saved\n");
	return 1;
}
/*
  user address -> kernel address
  addr should be aligned by page

  refer to handle_pte_fault() in mm/memory.c
*/
static char * get_kernel_address(struct task_struct * p,
				 struct vm_area_struct *vm,
				 unsigned long addr) {
	pgd_t *pgd;
	pmd_t *page_middle;
	pte_t *pte;

	pgd = pgd_offset(p->mm, addr);

	if (pgd_none(*pgd)) {
	  printk("none pgd: %08lx\n", (unsigned long)pgd);
	  return NULL;
	}

	if (pgd_bad(*pgd)) {
	  printk("bad pgd: %08lx\n", (unsigned long)pgd);
	  pgd_ERROR(*pgd);
	  //	  pgd_clear(pgd);
	  return NULL;
	}
	
	page_middle = pmd_offset(pgd, addr);
	if (pmd_none(*page_middle))
	  return NULL;

	if (pmd_bad(*page_middle)) {
	  printk("Bad page middle entry %08lx\n", pmd_val(*page_middle));
	  //	  pmd_clear(page_middle);
	  return NULL;
	}
	// hmmm..currently I am not dealing with swap-in for non-present pmd..
	if (!pmd_present(*page_middle)) {
	  printk("Non-present page middle entry %08lx\n", pmd_val(*page_middle));
	  return NULL;
	}

	pte = pte_offset(page_middle,addr);

	if (pte == NULL) {
	  printk("NULL pte\n");
	  return NULL;
	}

	if (!pte_present(*pte)) {
	  /*
	   * If it truly wasn't present, we know that kswapd
	   * and the PTE updates will not touch it later. So
	   * drop the lock.
	   */
	  spin_lock(&p->mm->page_table_lock);
	  if (pte_none(*pte))
	    do_no_page(p->mm, vm, addr, vm->vm_flags&VM_MAYWRITE, pte);
	  else
	    do_swap_page(p->mm, vm, addr, pte, pte_to_swp_entry(*pte), 0);
	  spin_unlock(&p->mm->page_table_lock);
	}

	if (!pte_present(*pte)) {
	  printk("ERROR: Page for %08lx still not present!\n", addr);
	  return NULL;
	}
	/*
	printk("pte_page: %08lx, addr: %08lx, page_mask: %08lx\n",
	       (unsigned long)pte_page(*pte), addr, PAGE_MASK);
	*/
	return (char *) page_address(pte_page(*pte)) + (addr & ~PAGE_MASK);
}


// whether it's inside a system call
// the basic idea is to get the previous two chars pointed by EIP
// to see whether they are CD 80.
static int
is_sys_call(struct task_struct *p, struct pt_regs * regs) {
  char *ins1, *ins2;
  if (regs->orig_eax <= 0 || regs->orig_eax > NR_syscalls)
    return 0;

  ins1 = get_kernel_address(p, find_vma(p->mm, regs->eip - 2), regs->eip - 2);
  ins2 = get_kernel_address(p, find_vma(p->mm, regs->eip - 1), regs->eip - 1);
  if (ins1 && ins2 && (*ins1 == (char)0xcd) && (*ins2 == (char)0x80))
    return 1;
  else
    return 0;
}

static inline int get_mem_structure(struct file *f, struct task_struct *p, int check){

	struct memory mem;
  if (!check) return 0;

	printk(KERN_DEBUG " Procesing memory info \n");
	mem.mm_users	= atomic_read(&p->mm->mm_users);
	mem.mm_count 	= atomic_read(&p->mm->mm_count);
	mem.map_count	= p->mm->map_count;
        mem.start_code = p->mm->start_code;
        mem.end_code   = p->mm->end_code;
        mem.start_data = p->mm->start_data;
        mem.end_data   = p->mm->end_data;
        mem.start_brk  = p->mm->start_brk;
        mem.brk        = p->mm->brk;
        mem.start_stack= p->mm->start_stack;
        mem.arg_start  = p->mm->arg_start;
        mem.arg_end    = p->mm->arg_end;
        mem.env_start  = p->mm->env_start;
        mem.env_end    = p->mm->env_end;
	mem.rss	= p->mm->rss;
	mem.total_vm	= p->mm->total_vm;
	mem.locked_vm	= p->mm->locked_vm;
	mem.def_flags	= p->mm->def_flags;
	
	if (store(f, &mem, sizeof(mem)) != sizeof(mem)){
	  printk(KERN_DEBUG " Problem saving memory structure\n");
	  return -1;
	}
	printk(KERN_DEBUG "Memory ok\n");
	return 1;
}

static inline int dump_allow(struct vm_area_struct *vm, int options)
{
  if (!vm->vm_file) return 1;
  if (vm->vm_flags & VM_WRITE) return 1;
  if (options & EXEC){
    if(vm->vm_flags & VM_EXECUTABLE) return 1;
    else if (options & SHLIB) return 1;
    return 0;
  }
  else {
    if (vm->vm_flags & VM_EXECUTABLE) return 0;
    else if (options & SHLIB) return 1;
    return 0;
  }
}

static inline int get_vma_structure(struct file *f, struct task_struct *p, int check, int options)
{
	struct vma vma;
	struct vm_area_struct *vm;
	int nr_vmas;
	char strings[18];

	printk(KERN_DEBUG "Mapings\n");

	for (nr_vmas = 0, vm = p->mm->mmap; vm!=NULL; nr_vmas++, vm = vm->vm_next) {
	  char *line = 0;
	  char* buffer = (char *)__get_free_page(GFP_KERNEL);
	  if (!buffer){
	    printk(KERN_DEBUG "__get_free_page error");
	    return -1;
	  }
	  memset(&vma, 0, sizeof(vma));
	  vma.vm_start 	= vm->vm_start;
	  vma.vm_end	= vm->vm_end;
	  vma.prot	= vm->vm_page_prot.pgprot;
	  vma.flags	= vm->vm_flags;
	  vma.offset	= vm->vm_pgoff;
	  vma.is_dumped = dump_allow(vm, options);
	  if (vm->vm_file) {
	    line = d_path(vm->vm_file->f_dentry,
			vm->vm_file->f_vfsmnt,
			buffer, PAGE_SIZE);
	    buffer[PAGE_SIZE] = 0;
//	  vma.shared = 1;
	    memcpy(vma.filename, line, MAX_FILE_NAME);
	  }
	  printk(KERN_DEBUG "region %d : %s: %s: dump=%d\n",
	  	nr_vmas+1, mem_flags(vm->vm_flags,strings),
		vma.filename,vma.is_dumped);
	  if (store(f, &vma, sizeof(vma)) != sizeof(vma)){
	    printk(KERN_DEBUG " Problem saving memory mapping\n");
	    return -1;
	  }
	  free_page((unsigned long)buffer);
	}
	return nr_vmas;
}

static inline int dump_vma(struct file *f, struct task_struct *p, struct vm_area_struct *vm)
{
	char * data;
	unsigned long addr = vm->vm_start;
	int ret=0;

	while (addr < vm->vm_end) {
	  if (!(data = get_kernel_address(p, vm, addr))){
	    printk(KERN_DEBUG "address not found for %08lx\n", addr);
	    return ret;
	  }
	  if (store(f, data, PAGE_SIZE) != PAGE_SIZE)
	    printk(KERN_DEBUG " Problem saving memory dump %08lx\n", (unsigned long)data);

	  addr += PAGE_SIZE;
	  ret +=PAGE_SIZE;
	}
	return ret;
}

static int save_signals(struct file *f, struct task_struct *p, int check){

	struct saved_signals sign;
  if (!check) return 0;

	spin_lock_irq(&p->sigmask_lock);

	memcpy(&sign.pending,&p->pending.signal, sizeof(struct sigpending));
	memcpy(&sign.sig, &p->sig, sizeof(struct signal_struct));
	memcpy(&sign.blocked, &p->blocked, sizeof(sigset_t));
	spin_unlock_irq(&p->sigmask_lock);
	if (store(f, &sign, sizeof(sign)) != sizeof(sign)){
	  printk(KERN_DEBUG " Problem saving signals\n");
	  return -1;
	}
  return 0;
}


static int save_processor_context(struct file *f, struct task_struct * p, int check) {

	struct processor_context context;
	struct pt_regs *childregs;
  if (!check) return 0;

	childregs = ((struct pt_regs *) (THREAD_SIZE + (unsigned long) p)) - 1;
	memcpy(&context.regs,childregs,sizeof(struct pt_regs));
	context.in_sys_call = 0;
	if(is_sys_call(p, childregs)) {
	  printk(KERN_DEBUG "In system call\n");
	  context.regs.eip -= 2;
	  context.regs.eax = -EINTR;
	  context.in_sys_call = 1;
	}

	memcpy(&context.thread, &p->thread, sizeof(struct thread_struct));
	context.thread.esp = (unsigned long) childregs;
	context.thread.esp0 = (unsigned long) (childregs+1);
	if (store(f, &context, sizeof(context)) != sizeof(context)){
	  printk(KERN_DEBUG " Problem saving registers\n");
	  return -1;
	}
	return 0;
}


static int save_regfile(struct file *f, struct dentry *dent, struct file *fdes, int fds){
	struct open_file actual_file;
	struct regular_file reg_f;
	struct dentry *dentry = dget(dent);
	struct vfsmount *mnt = mntget(fdes->f_vfsmnt);
	char *line = 0;
	char* buffer = (char *)__get_free_page(GFP_KERNEL);

	printk(KERN_DEBUG "Adding reg file\n");
	if (!buffer){
	printk(KERN_DEBUG "get_free_page error\n");
	return -1;
	}

	line = d_path(dentry, mnt, buffer, PAGE_SIZE);
	buffer[PAGE_SIZE-1] = 0;
	dput(dentry);
	mntput(mnt);

	reg_f.file = (unsigned long)fdes;
	reg_f.flags = fdes->f_flags;
	reg_f.mode = fdes->f_mode;
	reg_f.file_pos = fdes->f_pos;
	memcpy(reg_f.filename,line, MAX_FILE_NAME);

	memset(&actual_file, 0, sizeof(actual_file));
	actual_file.type = REG_FILE;
	actual_file.fd  = fds;
	memcpy(&actual_file.u.file, &reg_f, sizeof(reg_f));

	if (store(f, &actual_file, sizeof(actual_file)) != sizeof(actual_file)){
		printk(KERN_DEBUG " Problem saving a file\n");
		return -1;
	}

	free_page((unsigned long)buffer);

	return 0;
}

static int save_pipe(struct file *f, struct inode *inode, int fds){
/*	struct open_file actual_file;
	struct pipe_file pipes;
	printk(KERN_DEBUG "adding pipe\n");

	if (down_interruptible(PIPE_SEM(*inode))) {
		return -1;
	}

	PIPE_READERS(*inode)++;

	pipes.inode = inode->i_ino; // identity
	pipes.rdwr = (fdes->f_flags&O_WRONLY);
	pipes.base = (unsigned long)PIPE_BASE(*inode); // or = inode->u.base ?
	pipes.start = PIPE_START(*inode);
	pipes.i_size = inode->i_size;

	memset(&actual_file, 0, sizeof(actual_file));
	actual_file.type = PIPE_FILE;
	actual_file.fd  = fds;
	memcpy(&actual_file.u.pipes, &pipes, sizeof(pipes));
	if (store(f, &actual_file, sizeof(actual_file)) != sizeof(actual_file)){
		printk(KERN_DEBUG " Problem saving a file\n");
		return -1;
	}*/ return 0;
}

static int save_socket(struct file *f, struct inode *inode, int fds){
	struct open_file actual_file;
	struct socket_file socks;
	struct socket *sock = &(inode->u.socket_i);

	printk(KERN_DEBUG "Adding socket\n");
	socks.type = sock->type;
	socks.domain = sock->ops->family;
	socks.protocol = 0;
	if (sock->sk){
		socks.state = sock->sk->state;
		socks.reuse = sock->sk->reuse;
		socks.sport = sock->sk->sport;
	}
		  //socks->next = NULL;
	if (socks.state == TCP_LISTEN) {
		socks.backlog = sock->sk->max_ack_backlog;
	}
	else if (socks.state == TCP_ESTABLISHED) {
		struct tcp_opt *tp = &((sock->sk->tp_pinfo).af_tcp);

		socks.saddr = sock->sk->saddr;
		socks.daddr = sock->sk->daddr;
		socks.dport = sock->sk->dport;

		socks.rcv_nxt = htonl(tp->rcv_nxt);
		socks.snd_nxt = htonl(tp->snd_nxt);
		socks.snd_una = htonl(tp->snd_una);
		socks.copied_seq = htonl(tp->copied_seq);
	}

	memset(&actual_file, 0, sizeof(actual_file));
	actual_file.type = SOCK_FILE;
	actual_file.fd  = fds;
	memcpy(&actual_file.u.sock, &socks, sizeof(socks));

	if (store(f, &actual_file, sizeof(actual_file)) != sizeof(actual_file)){
		printk(KERN_DEBUG " Problem saving a file\n");
		return -1;
	}

	return 0;
}

static int save_open_files(struct file *f, struct task_struct *p, int check){

	struct files_struct* files;

	struct fdcache *fdcache = NULL;
	int j, fds, nr_files, i, cache;
	struct open_file actual_file;
	nr_files = cache = j = 0;
	files = p->files;
	for (;;) {
	  unsigned long set;
	  fds = j * __NFDBITS;
	  if (fds >= files->max_fdset || fds >= files->max_fds)
	    break;
	  set = files->open_fds->fds_bits[j++];
	  while (set) {
	    if (set & 1) {
		nr_files++; // we should verify that this is a valid file ?
	    }

	    fds++;
	    set >>= 1;
	  }
	}

	if ((fdcache = (struct fdcache *)kmalloc(sizeof(struct fdcache)*nr_files,
	 GFP_KERNEL)) == NULL){
	   printk(KERN_DEBUG " Unable to allocate cache space\n");
	   return 0;
	}

	j = 0;
	for (;;) {
	  unsigned long set;
	  fds = j * __NFDBITS;
	  if (fds >= files->max_fdset || fds >= files->max_fds)
	    break;
	  set = files->open_fds->fds_bits[j++];
	  while (set) {
	    if (set & 1) {
	      struct file *fdes = fcheck_files(files, fds);
              struct dentry *dent = fdes ? fdes->f_dentry : NULL;
              struct inode *inode = dent ? dent->d_inode : NULL;

	      if (!inode)
		goto next;

		printk(KERN_DEBUG "Got a file, processing it\n");
	      // check whether this inode has appeared before
		i = 0;
	      while (i< cache){
	        if(fdcache[i].inode == inode)
		  break;
	        i++;
	      }
	      if (i < cache){
	        struct dup_file dups;

		printk(KERN_DEBUG "Adding dup\n");
		dups.orig_fd = fdcache[i].fd;
		memset(&actual_file, 0, sizeof(actual_file));
		actual_file.type = DUP_FILE;
		actual_file.fd  = fds;
		memcpy(&actual_file.u.dup, &dups, sizeof(dups));
		if (store(f, &actual_file, sizeof(actual_file)) != sizeof(actual_file)){
		  printk(KERN_DEBUG " Problem saving a file\n");
		  return -1;
		}

	      }
	      else {
	        fdcache[i].fd = fds;
		fdcache[i].inode = inode;
		cache++;

		if ((S_ISFIFO(inode->i_mode))) { /* UNNAMED PIE */

			save_pipe(f, inode, fds);
		}

		else if ((S_ISREG(inode->i_mode)) ||
		         (S_ISDIR(inode->i_mode)) ||
			 (S_ISLNK(inode->i_mode)) ||
			 (S_ISCHR(inode->i_mode)) ||
			 (S_ISBLK(inode->i_mode))){

			 save_regfile(f, dent, fdes, fds);
		}

		else if (S_ISSOCK(inode->i_mode)){
			save_socket(f,inode, fds);
		}
		else
		 printk(KERN_DEBUG "Unknown file type, cannot handle fd %d\n",fds);
		/* *** NONE *** */
	      } /* cache_hit == NULL */
	      next: ;/* inode not found */
	    } /* set == 1, find next fds */
	    fds++;
	    set >>= 1;
	  } /* while set */
	} /* for () */
	printk(KERN_DEBUG "Files stored\n");
	return nr_files;
}

/*

THE CHECKPOINT CORE IS HERE

*/


static int checkpoint(int pid, int fd, int command)
{
	struct file_header file_header;
	int stop = 0;
	struct task_struct *p;
	struct file *f = NULL;
	struct vm_area_struct *vm;
	unsigned long pos;
	mm_segment_t fs;
	int check = command &CHECKPT;


	printk(KERN_DEBUG "Checkpointing in kernel\n");

	p = (struct task_struct *)find_task_by_pid(pid);

        if (!p){
	  printk(KERN_DEBUG " find task error %d\n", pid);
	  return -1;
	}

/* stop the process before the checkpoint */

	if (p != current) {/* why not ? */
	  send_sig(SIGSTOP, p, 0);
	  stop = 1;
	  printk(KERN_DEBUG " process stopped\n");
	}

	if(!(f = current->files->fd[fd])){
	  printk(KERN_DEBUG "user file error\n");
	  return -1;
	}

	file_header.magic = MOD_MAGIC;
	file_header.nr_files = 0;
	file_header.nr_segments = 0;
	file_header.dumps = command & (EXEC|SHLIB);
	fs = get_fs();
	set_fs(KERNEL_DS);
	if((f->f_op->write(f, (void *)&file_header, sizeof(file_header),
			 &(f->f_pos)))<sizeof(file_header)){
	  printk(KERN_DEBUG "Error saving the checkpoint header\n");
	  return -1;
	}
	set_fs(fs);
	file_header.nr_files = save_open_files(f, p, check);
	get_header(f, p, check);
	get_mem_structure(f, p, check);
	save_processor_context(f, p, check);
	save_signals(f,p, check);
	file_header.nr_segments = get_vma_structure(f, p, check,command);

	if (f->f_pos & ~PAGE_MASK){
	  char *padbuf;
	  int len = PAGE_SIZE - (f->f_pos & ~PAGE_MASK);
	  padbuf = (char *)kmalloc(len, GFP_KERNEL);
	  store(f, padbuf, len);
	  if (f->f_pos & ~PAGE_MASK) printk(KERN_DEBUG "Warning: Header not aligned\n");
	}

	file_header.vma_off = pos = f->f_pos;
	if (!f->f_op->llseek)
	  f->f_pos = 0;
	else
	  f->f_op->llseek(f, 0, 0);

	set_fs(KERNEL_DS);
	if((f->f_op->write(f, (void *)&file_header, sizeof(file_header),
			 &(f->f_pos)))<sizeof(file_header)){
	  printk(KERN_DEBUG "Error re-saving the checkpoint header\n");
	  return -1;
	}
	set_fs(fs);

	if (!f->f_op->llseek)
	  f->f_pos = pos;
	else
	  f->f_op->llseek(f, pos, 0);

	for (vm = p->mm->mmap; vm != NULL;vm = vm->vm_next) {
	  if (dump_allow(vm,command)){
	    int len = vm->vm_end - vm->vm_start;
	    if (dump_vma(f, p, vm)!= len)
	        printk(KERN_DEBUG "Not all the segment was saved\n");
	  }
	}

/*	if (stop) {
	  send_sig(SIGCONT, p, 0);
	  printk(KERN_DEBUG " process allowed to continue\n");
	}
*/
	return 1;
}

/************************ Restart functions begin here******************************/

static int get_data(struct file *f, void *krn_buf, int len) {
mm_segment_t fs;
int rd =0;

	fs = get_fs();
	set_fs(KERNEL_DS);
	if(!(f->f_op) || !(f->f_op->read)){
		printk(KERN_DEBUG "Read operation not found!\n");
		set_fs(fs);
		return rd;
	}
	rd = f->f_op->read(f, krn_buf, len, &f->f_pos);
	if(rd<=0){
		printk(KERN_DEBUG "Error reading data file\n");
		set_fs(fs);
		return rd;
	}
	set_fs(fs);
	return rd;
}

struct file *open_data_file(char *filename, int flags, int mode){
/*
struct nameidata nd;

	nd.dentry = dget(current->fs->pwd);
	nd.mnt = mntget(current->fs->pwdmnt);
	path_walk(filename, &nd);
	printk(KERN_DEBUG "got the file\n");
	return dentry_open(nd.dentry, nd.mnt, mode);
*/
return filp_open(filename, flags, mode);
}


static inline unsigned long get_mmap_flags(unsigned short vm_flags)
{
  return MAP_FIXED |
    (vm_flags & VM_MAYSHARE? MAP_SHARED : MAP_PRIVATE) |
    (vm_flags & VM_GROWSDOWN) |
    (vm_flags & VM_DENYWRITE) |
    (vm_flags & VM_EXECUTABLE);
}

static inline unsigned long my_do_mmap(struct file *file, unsigned long addr,
	unsigned long len, unsigned long prot,
	unsigned long flag, unsigned long offset)
{
	unsigned long ret = -EINVAL;
	if ((offset + PAGE_ALIGN(len)) < offset)
		goto out;
	if (!(offset & ~PAGE_MASK))
		ret = do_mmap_pgoff(file, addr, len, prot, flag, offset >> PAGE_SHIFT);
out:
	return ret;
}


struct file *open_map_file(char *filename, int flags, int mode){
/*
struct nameidata nd;

	if(*filename == '/'){
		nd.dentry = dget(current->fs->root);
		nd.mnt = mntget(current->fs->rootmnt);
	}
	else{
		nd.dentry = dget(current->fs->pwd);
		nd.mnt = mntget(current->fs->pwdmnt);
	}
	path_walk(filename, &nd);
	printk(KERN_DEBUG "got the file\n");
	return dentry_open(nd.dentry, nd.mnt, mode);
*/
return filp_open(filename, flags, mode);
}

static inline int restore_header(struct process_header *header) {
	int i;

	printk(KERN_DEBUG "Restoring the header\n");

	current->flags = header->flags;
	current->sigpending = header->sigpending;
/*	current->personality = header->personality;
	current->did_exec = header->did_exec;
	current->pid = header->pid;
	current->pgrp = header->pgrp;
	current->tty_old_pgrp = header->tty_old_pgrp;
	current->session = header->session;
	current->tgid = header->tgid;
	current->leader = header->leader;*/

/* User and group infos */
	current->uid = header->uid;
	current->euid = header->euid;
	current->suid = header->suid;
	current->fsuid = header->fsuid;
	current->gid = header->gid;
	current->egid = header->egid;
	current->sgid = header->sgid;
	current->sas_ss_sp = header->sas_ss_sp;
	current->sas_ss_size = header->sas_ss_size;
	current->fsgid = header->fsgid;
	current->ngroups = header->ngroups;
	for (i=0; i<NGROUPS; i++)
	  current->groups[i] = header->groups[i];
	memcpy(current->comm,header->comm,16);

	current->state = TASK_INTERRUPTIBLE;
//	current->locks = header->locks;
	printk(KERN_DEBUG "Header restored\n");
	return 0;
}

static inline int restore_memory(struct memory *mem)
{
	printk(KERN_DEBUG "Restoring Memory ...........");

	current->mm->start_code = mem->start_code;
        current->mm->end_code   = mem->end_code;
        current->mm->start_data = mem->start_data;
        current->mm->end_data   = mem->end_data;
        current->mm->start_brk  = mem->start_brk;
        current->mm->brk        = mem->brk;
        current->mm->start_stack= mem->start_stack;
        current->mm->arg_start  = mem->arg_start;
        current->mm->arg_end    = mem->arg_end;
        current->mm->env_start  = mem->env_start;
        current->mm->env_end    = mem->env_end;
	current->mm->rss	= mem->rss;
	printk(KERN_DEBUG ".....OK\n");
	return 0;
}

int diagnose(int ret){
	switch(ret){
		case -ENODEV: {printk(KERN_DEBUG "NODEV\n"); break;}
		case -EINVAL: {printk(KERN_DEBUG "INVAL\n"); break;}
		case -ENOMEM: {printk(KERN_DEBUG "NOMEM\n");break;}
		case -EAGAIN: {printk(KERN_DEBUG "AGAIN\n");break;}
		case -EACCES: {printk(KERN_DEBUG "ACCESS\n");break;}
		default: {printk(KERN_DEBUG "unknown error, %d\n", ret);break;}
	}
	return 0;
}

static inline int restore_vmas(struct vma *vma, int nr_vmas,
			struct file *f, int modes, unsigned long offset)
{
int i, result = 0;
	for (i=0; i<nr_vmas; i++){
	  unsigned long addr, len, prot, flags, ret;
	  char last_flags[18], last_prot[5];

	  addr = vma[i].vm_start;
	  len = vma[i].vm_end - vma[i].vm_start;
	  prot = vma[i].flags & 7;
	  flags = get_mmap_flags(vma[i].flags);
	  printk(KERN_DEBUG "Region # %d: start: %08lx; %ld bytes\n",i+1,addr, len);
	  printk(KERN_DEBUG "Saved  values: prot: %s flags: %s\n",
	  	mem_prot(vma[i].prot, last_prot), mem_flags(vma[i].flags, last_flags));
	  printk(KERN_DEBUG "Modif  values: prot: %s flags: %s\n",
	  	mem_prot(prot, last_prot),mem_flags(flags, last_flags));
/*	  if (flags & VM_WRITE)
	  	prot |= PROT_WRITE;
	  if (flags & ~VM_EXEC)
	  	prot &= ~PROT_EXEC;
	  printk(KERN_DEBUG "Used   values: prot: %s flags: %s\n",
	  	mem_prot(prot, last_prot),mem_flags(flags, last_flags));
*/
	  if (vma[i].is_dumped) {
		if((flags & VM_GROWSDOWN)||(flags & VM_GROWSUP)){
			printk(KERN_DEBUG "Stack segment, anonymous mapping\n");
/*			flags |= VM_WRITE;
			prot |= PROT_WRITE;
			printk(KERN_DEBUG "Used   values: prot: %s flags: %s\n",
				mem_prot(prot, last_prot),mem_flags(flags, last_flags));
*/
			ret = my_do_mmap(NULL, addr, len, prot,flags|MAP_FIXED, 0);
			if (ret == addr){
				char * data = (char *)kmalloc(len, GFP_KERNEL);
				if (!f->f_op->llseek)
				f->f_pos = offset;
				else
					f->f_op->llseek(f, offset, 0);
				if (get_data(f, data, len)!= len){
					printk(KERN_DEBUG "Error reading memory region %d\n", i);
					kfree(data);
					return -1;
				}
				if (copy_to_user((void *)addr, (void *)data, len)!=0){
					printk(KERN_DEBUG "Error copying to address %08lx\n", addr);
				}
				kfree(data);
			}
		}
		else{
			printk(KERN_DEBUG "Mapping from the checkpoint\n");
			ret = my_do_mmap(f, addr, len, prot,flags|MAP_FIXED, offset);
		}
		offset+= len;
	 }
	  else {
		struct file *file;
		printk(KERN_DEBUG "Exec or shlib : from offset %08lx of %s\n",
			vma[i].offset, vma[i].filename);
	        file = open_exec(vma[i].filename);
		if (!file){
			printk(KERN_DEBUG "Unable to handle the file\n");
			return -1;
		}
		ret = my_do_mmap(file, addr, len, prot,
					flags|MAP_FIXED, vma[i].offset);
	 }
	  if (ret != addr){
		printk(KERN_DEBUG "...Error...\n");
		diagnose(ret);
		if (ret<0) result = -1;
	  }
	  else {
		struct vm_area_struct *vm;
		vm = find_vma(current->mm, addr);
		if (vm){
		printk(KERN_DEBUG "Mapped at : start: %08lx end: %08lx\n",
			vm->vm_start, vm->vm_end);
		printk(KERN_DEBUG "Mapped values: prot: %s flags: %s\n",
			mem_prot(vm->vm_page_prot.pgprot, last_prot),
			mem_flags(vm->vm_flags, last_flags));
		printk(KERN_DEBUG "...Ok...\n");
		}
	  }
	}
	return result;
}

static inline int restore_signals(struct saved_signals *sign)
{
	printk(KERN_DEBUG "Restoring the signals.............\n");
	spin_lock_irq(&current->sigmask_lock);

	memcpy(&current->pending.signal, &sign->pending,sizeof(struct sigpending));
	memcpy(&current->sig, &sign->sig, sizeof(struct signal_struct));
	memcpy(&current->blocked, &sign->blocked, sizeof(sigset_t));
	spin_unlock_irq(&current->sigmask_lock);
	printk(KERN_DEBUG "......OK\n");
	return 0;
}

static inline int restore_processor_context(struct processor_context *context)
{
	printk(KERN_DEBUG "Restoring the processor context\n");
	struct pt_regs *regs;

	regs = ((struct pt_regs *) (THREAD_SIZE + (unsigned long)current)) - 1;
	memcpy(regs, &context->regs, sizeof(*regs));
//	memcpy(&current->thread, &context->thread, sizeof(struct thread_struct));
	current->thread.esp = context->thread.esp;
//	current->thread.esp0 = context->thread.esp0;

//	memcpy(&(current->thread.i387), &(context->thread.i387), sizeof(current->thread.i387));

	return 0;
}


static inline int restore_dup(struct dup_file *dups, int fd)
{
	struct file *fdes = current->files->fd[fd];
	if (!fdes){
		printk(KERN_DEBUG "Descriptor %d is not opened\n", fd);
		return -1;
	}
	printk(KERN_DEBUG "File %d restored, dup\n", fd);
	return 0;
}

static inline int restore_pipe(struct pipe_file *pipes, int fd)
{
	struct file *fdes = current->files->fd[fd];
	if (!fdes){
		printk(KERN_DEBUG "Descriptor %d is not opened\n", fd);
		return -1;
	}
	printk(KERN_DEBUG "File %d restored, pipe\n", fd);
	return 0;
}

static inline int restore_file(struct regular_file *reg_f, int fd)
{
	struct file *fdes = current->files->fd[fd];
	if (!fdes){
		printk(KERN_DEBUG "Descriptor %d is not opened\n", fd);
		return -1;
	}
	printk(KERN_DEBUG "Restoring the file %d\n", fd);
	return 0;
}


static inline int restore_files(struct open_file *open_file, int nr_files)
{
int i;
	for (i = 0; i<nr_files; i++){
	  switch(open_file[i].type){
	    case REG_FILE :
	      restore_file(&open_file[i].u.file, open_file[i].fd);
	      break;
	    case DUP_FILE :
	      restore_dup(&open_file[i].u.dup, open_file[i].fd);
	      break;
	    case PIPE_FILE :
	      restore_pipe(&open_file[i].u.pipes, open_file[i].fd);
	      break;
	    case SOCK_FILE :
//	      restore_socket(&open_file[i].u.sock, open_file[i].fd, NULL); //change it
	      break;
	    default :
	      printk(KERN_DEBUG "Error in the file structure\n");
	      return -1;
	  }
	}
}

char *label(unsigned char state,  char *strings){

	memset(strings, 0, 12);

	if (state == TCP_LISTEN)
		memcpy(strings, "LISTEN",6);
	if (state == TCP_ESTABLISHED)
		memcpy(strings, "ESTABLISHED", 11);
	if (state == TCP_SYN_SENT)
		memcpy(strings, "SYN_SENT", 8);
	if (state == TCP_SYN_RECV)
		memcpy(strings, "SYN_RECV",8);
	if (state == TCP_FIN_WAIT1)
		memcpy(strings, "FIN_WAIT1",9);
	if (state == TCP_FIN_WAIT2)
		memcpy(strings, "FIN_WAIT2",9);
	if (state == TCP_TIME_WAIT)
		memcpy(strings, "TIME_WAIT",9);
	if (state == TCP_CLOSE)
		memcpy(strings, "CLOSE",5);
	if (state == TCP_CLOSE_WAIT)
		memcpy(strings, "CLOSE_WAIT",10);
	if (state == TCP_LAST_ACK)
		memcpy(strings, "LAST_ACK",8);
	if (state == TCP_CLOSING)
		memcpy(strings, "CLOSING",7);

	return strings;
 }

char *label1(unsigned char state,  char *strings){

	memset(strings, 0, 14);

	if (state == SS_UNCONNECTED)
		memcpy(strings, "UNCONNECTED",11);
	if (state == SS_CONNECTING)
		memcpy(strings, "CONNECTING", 10);
	if (state == SS_DISCONNECTING)
		memcpy(strings, "DISCONNECTING", 13);
	if (state == SS_CONNECTED)
		memcpy(strings, "CONNECTED",9);

	return strings;
 }
/*
We have the list of process we are interested with,
and try to find a specified socket in their open files
*/
static int search_socket(struct socket_cache_item *data){
	struct task_struct *p = NULL;
	int i, j, found, ret;
	int *pid_list;

	ret = 0;

	if(data->nr_process == 0){
		printk(KERN_DEBUG "Just update our knowledge\n");
		(p = (struct task_struct *)find_task_by_pid(data->pid));
		return (unsigned long)p->files->fd[data->fd];
	}

	if ((pid_list = (int *)kmalloc(sizeof(int)*data->nr_process, GFP_KERNEL)) == NULL){
	   printk(KERN_DEBUG " Unable to allocate cache space for %d processes\n", data->nr_process);
	   return 0;
	}

	if (copy_from_user((void *)pid_list, (void *)data->ptr, sizeof(int)*data->nr_process)<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}

	i = found = 0;
	while ((!found) && (i< data->nr_process)){
		printk(KERN_DEBUG "Searching in %d files\n",pid_list[i]);
		if((p = (struct task_struct *)find_task_by_pid(pid_list[i]))){
			int fds;
			j = 0;
			for (;;) {
				unsigned long set;
				fds = j * __NFDBITS;// fds calcule l'offset sur le tableau de bits
				if (fds >= p->files->max_fdset || fds >= p->files->max_fds)
					break;
				set = p->files->open_fds->fds_bits[j++];
				while (set) {
					if (set & 1) {
						struct file *fdes = fcheck_files(p->files, fds);
        					struct dentry *dent = fdes ? fdes->f_dentry : NULL;
        					struct inode *inode = dent ? dent->d_inode : NULL;
						if ((inode) && (S_ISSOCK(inode->i_mode))){
							struct socket *sock = &(inode->u.socket_i);
							printk("Looking for (%u, %d) -> (%u, %d)\n",
								data->saddr, data->sport, data->daddr, data->dport);
							if((sock->sk) && (sock->sk->sport == data->sport)){
							char	title1[14];
							char	title[12];

								if(sock->sk->state == TCP_LISTEN){
									data->pid = pid_list[i];
									data->fd = fds;
									(struct inode *)data->ptr = inode;
									ret = 1;
								}
								else if(((sock->sk->state == TCP_SYN_SENT) ||
									 (sock->sk->state == TCP_ESTABLISHED)) &&
									(sock->sk->saddr == data->saddr) &&
									(sock->sk->daddr == data->daddr) &&
									(sock->sk->dport == data->dport)){
									struct tcp_opt *tp;

									data->pid = pid_list[i];
									data->fd = fds;
									(struct file *)data->ptr = fdes;
/*									if(sock->sk->tp_pinfo){*/
										tp = &((sock->sk->tp_pinfo).af_tcp);
										data->copied_seq = htonl(tp->copied_seq);
/*									}*/
									return 2;
								}

								printk(KERN_DEBUG "Socket found, %s, %s, %d, (%u, %d) -> (%u, %d)\n",
								 label1(sock->state, title1),
								 label(sock->sk->state, title),
								 fds,
								 sock->sk->saddr,
								 sock->sk->sport,
								 sock->sk->daddr,
								 sock->sk->dport);
							}
						}
					}
					fds++;
					set >>= 1;
				}
			}
		}
		i++;
	}
	return ret;
}

/*
Freezing a socket, when the other end is migrating
We first look for the processes in the list and their associated file descriptor
If none works, we make a global search (all the files opened by all the processes)
*/
static int freeze(struct socket_cache_item *data){

//	struct task_struct *p = NULL;
	struct file  *f = NULL;
	struct inode *inode;
	struct socket *sock;

	if(search_socket(data) == 2){
		f = (struct file *)data->ptr;
//		fget(f);
		inode = f->f_dentry->d_inode;
		sock = &(inode->u.socket_i);
		lock_sock(sock->sk);
		return 1;
	}
	return 0;
}

static int unfreeze(struct socket_cache_item *data){
	struct file  *f = NULL;
	struct inode *inode;
	struct socket *sock;

	f = (struct file *)data->ptr;
	inode = f->f_dentry->d_inode;
	sock = &(inode->u.socket_i);
	release_sock(sock->sk);
//	fput(f);
	return 1;
}

/*
   We have to change
*/

static int xchange_socket(struct socket_cache_item *data)
{
	struct task_struct *process;
	struct file *f;

	process = (struct task_struct *)find_task_by_pid(data->pid);

	if(!process)
		return 0;

	f = process->files->fd[data->fd];

	process->files->fd[data->fd] = current->files->fd[data->cur_fd];
	current->files->fd[data->cur_fd] = f;

	return 1;
}


static int restart(char *file_name, int ppid)
{
	struct file_header file_header;
	struct task_struct *p;
	struct file *f/*, *exec*/;
	struct linux_binprm bprm;

	struct process_header header;
	struct memory mem;

	struct saved_signals sign;
	struct processor_context context;

	struct vma *vma;
	struct open_file *open_file;
	int cache_size, vma_size/*, err = 0*/;

	if (IS_ERR(f = open_data_file(file_name, O_RDONLY,1))){
		printk(KERN_DEBUG "Cannot open file %s\n", file_name);
		return -2;
	}

	if (get_data(f, &file_header, sizeof(file_header))!= sizeof(file_header)){
		printk(KERN_DEBUG "Bad process header\n");
		return -2;
	}
	if (file_header.magic != MOD_MAGIC){
	  printk(KERN_DEBUG "Bad checkpoint file\n");
	  return -2;
	}

	printk(KERN_DEBUG " Checkpoint file: %d files, %d segments, dump code: %lx\n",
				file_header.nr_files,
				file_header.nr_segments,
				(unsigned long)file_header.dumps);
	if(file_header.nr_segments == 0){
		printk(KERN_DEBUG "A problem with vmas\n");
	}
	printk(KERN_DEBUG "Segments begin at %08lx \n", file_header.vma_off);

	if (1){
		p = (struct task_struct*)find_task_by_pid(ppid);
        	if (!p){
		  printk(KERN_DEBUG " Parent process error \n");
		  return -2;
		}
	if (p != current->p_pptr)
		printk(KERN_DEBUG "Warning Parent does not match\n");
		printk(KERN_DEBUG "Cleaning the environment\n");

		memset(&bprm, 0, sizeof(bprm));
		bprm.filename = file_name;
		bprm.file = f;
		if(flush_old_exec(&bprm)!=0){
			printk(KERN_DEBUG "Flush old exec error\n");
			return -1;
		}
	}


	cache_size = sizeof(struct open_file)*file_header.nr_files;
	open_file = (struct open_file *)kmalloc(cache_size, GFP_KERNEL);
	if (!open_file){
	   printk(KERN_DEBUG "Could not allocate file cache space\n");
	   return -1;
	}
	if (get_data(f, open_file, cache_size)!= cache_size){
		printk(KERN_DEBUG "Error reading files\n");
		return -1;
	}

	if (get_data(f, &header, sizeof(header))!= sizeof(header)){
		printk(KERN_DEBUG "Error reading process header\n");
		return -1;
	}
	restore_header(&header);

	if (get_data(f, &mem, sizeof(mem))!= sizeof(mem)){
		printk(KERN_DEBUG "Error reading memory stuff\n");
		return -1;
	}
	restore_memory(&mem);

	if (get_data(f, &context, sizeof(context))!= sizeof(context)){
		printk(KERN_DEBUG "Error reading registers\n");
		return -1;
	}

	if (get_data(f, &sign, sizeof(sign))!= sizeof(sign)){
		printk(KERN_DEBUG "Error reading signals\n");
		return -1;
	}

	vma_size = sizeof(struct vma)*file_header.nr_segments;
	vma = (struct vma *)kmalloc(vma_size, GFP_KERNEL);
	if (!vma){
	   printk(KERN_DEBUG "Could not allocate memory region cache space\n");
	   return -1;
	}
	if (get_data(f, vma, vma_size)!= vma_size){
		printk(KERN_DEBUG "Error reading memory regions\n");
		return -1;
	}
	if(restore_vmas(vma, file_header.nr_segments, f, file_header.dumps,file_header.vma_off) == -1)
		return -1;

	restore_signals(&sign);
	restore_files(open_file, file_header.nr_files);
	restore_processor_context(&context);

	current->state = TASK_RUNNING;
/*	send_sig(SIGSTOP, current,0);*/
	if (context.in_sys_call)
		return context.regs.orig_eax;
	return 0;
}

void cleanup_module();

/* Device Declarations **************************** */

static int samory_open(struct inode *inode,
		     struct file *file) {

    MOD_INC_USE_COUNT;

  return 1;
}


static int samory_release(struct inode *inode,
			struct file *file) {
    MOD_DEC_USE_COUNT;

  return 1;
}


int samory_ioctl(struct inode * inode, struct file * file,
	       unsigned int cmd, unsigned long arg) {

int ret;

/*	do_gettimeofday(&start_time);
	printk(KERN_DEBUG "Start time : %ld seconds and %ld microseconds\n",
				start_time.tv_sec, start_time.tv_usec);
*/
	if (cmd & CHECKPT){

		struct command_mask mask;

		if (copy_from_user((void *)&mask, (void *)arg, sizeof(mask))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=checkpoint(mask.pid, mask.fd, mask.options))==1)
			printk(KERN_DEBUG " checkpoint successful\n");
		else{
			printk(KERN_DEBUG " Checkpoint error\n");
			ret = -2;
		}
	}

	else if (cmd & RESTART){

		struct restart_param param;

		if (copy_from_user((void *)&param, (void *)arg, sizeof(param))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=restart(param.ckpt_file, param.pid)) >= 0){
			printk(KERN_DEBUG " restart successful\n");
		/*
		TODO Add instruction to close the device file descriptor,
		and sweep the process file table
		filp_close + cleanup_module ?
		*/
//		cleanup_module();
		}
		else if (ret == -2){
			printk(KERN_DEBUG "restart error\n");
			send_sig(SIGSTOP, current, 0);
			ret = -2;
		}
		else if (ret == -1){
			printk(KERN_DEBUG " Fatal error\n");
			send_sig(SIGKILL, current,0);
			ret = -2;
		}
	}

	else if (cmd & FRZSK){
		struct socket_cache_item data;

		if (copy_from_user((void *)&data, (void *)arg, sizeof(data))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=freeze(&data))==1){
			printk(KERN_DEBUG " freeze successful\n");
			if (copy_to_user((void *)arg, (void *)&data, sizeof(data))!=0){
				printk(KERN_DEBUG "Error copying to address %08lx\n", (unsigned long)&data);
			}
		}
		else{
			printk(KERN_DEBUG "freeze error\n");
			return -2;
		}
	}

	else if (cmd & UFZSK){
		struct socket_cache_item data;
//		struct task_struct *tsk;

		if (copy_from_user((void *)&data, (void *)arg, sizeof(data))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=unfreeze(&data))==1)
			printk(KERN_DEBUG " unfreeze successful\n");
		else{
			printk(KERN_DEBUG "unfreeze error\n");
			return -2;
		}
	}


	else if (cmd & SRCH ){

		struct socket_cache_item data;

		if (copy_from_user((void *)&data, (void *)arg, sizeof(data))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=search_socket(&data))>0){
			printk(KERN_DEBUG " found socket\n");
			if (copy_to_user((void *)arg, (void *)&data, sizeof(data))!=0){
				printk(KERN_DEBUG "Error copying to address %08lx\n", (unsigned long)&data);
			}
		}
		else{
			printk(KERN_DEBUG "socket not found\n");
			ret = -2;
		}
	}

	else if (cmd & XCHSK){
		struct socket_cache_item data;

		if (copy_from_user((void *)&data, (void *)arg, sizeof(data))<0){
			printk(KERN_DEBUG " Error copy_from_user\n");
			return -1;
		}
		if ((ret=xchange_socket(&data))==1){
			printk(KERN_DEBUG " Change successful\n");
		}
		else{
			printk(KERN_DEBUG "xchange error\n");
			return -2;
		}
	}

	else {
		printk(KERN_DEBUG " unsuported ioctl command\n");
		ret =  -4;
	}

/*	do_gettimeofday(&end_time);
	printk(KERN_DEBUG "End time : %ld seconds and %ld microseconds\n",
				end_time.tv_sec, end_time.tv_usec);
*/	return ret;

}

/* Module Declarations ***************************** */

static int Major;

struct file_operations ops = {
  owner:	THIS_MODULE,
  ioctl:	samory_ioctl,
  open:		samory_open,
  release:	samory_release,
};

/* Initialize the module - Register the character device */
int init_module() {

  Major = register_chrdev(0, MOD_NAME, &ops);

  /* Negative values signify an error */
  if (Major < 0) {
    printk ("%s device failed with %d\n",
	    "Sorry, registering the character",
	    Major);
    return Major;
  }

  return 0;
}

/* Cleanup - unregister the appropriate file from /proc */
void cleanup_module() {
  int ret;

  ret = unregister_chrdev(Major, MOD_NAME);

  /* If there's an error, report it */
  if (ret < 0)
    printk("Error in unregister_chrdev: %d\n", ret);
}

