#define DEVICE_NAME 	"/dev/samory"
#define MOD_NAME 	"samory"
#define MOD_MAGIC	876093
#define MOD_SERV_PORT	6421

#define MAX_FILE_NAME 	256
#define MAX_MSG_SIZE	65536
#define PACKET_SIZE	65536 /* for pack_write */
#define MAX_NR_SECTION  24 /* must be RLIMIT but this can change from a system to another */
#define MAX_PROCESS	8

#define FROM_KERNEL	0
#define FROM_USER	1

#define CKPT_T_OUT	300
#define PROP_T_OUT	180


/*
Sections headers (or types)
*/

#define HEADER		1
#define MEMORY		2
#define VMA		3
#define SIGNALS		4
#define REGISTERS	5
#define DUP_FILE	6
#define REG_FILE	7
#define PIPE_FILE	8
#define SOCK_FILE	9

/*
Options to the commands
*/
#define EXEC		1//shall we dump the executable?
#define SHLIB		2//shall we dump the shared libs?
#define LCFILE		4//do we keep the image in a local file ?
#define RRSTRT		8// should we restart in a remote host?
#define FAMILY		16 // checkpoint the hole family ?
#define HAVSKT		32 // must reopen the sockets

/*
Module command options
*/


#define CHECKPT		1  // checkpoint
#define RESTART		2  // restart
#define FRZSK		8  // freeze a socket
#define UFZSK		16 // release a frozen socket
#define SRCH		32 // tell if socket is to be monitored
#define XCHSK		64 // xchange socket after restart

struct samory_command{
	int type;
	int pid;
	char filename[MAX_FILE_NAME];
	char comm_file[MAX_FILE_NAME];
	char dest_host[16];
};

struct command_mask { /* this is were to store the dumps */
  int pid; /* process concerned */
  int fd; /* where to store the dumps */
  int options; /* family, alone, children, remote, shlib, exec... */
};

struct restart_param {
  int options;
  char ckpt_file[MAX_FILE_NAME];
  int fd;
  int pid;
};

struct socket_cache_item{
  void *ptr; // can be file address, or process table, or linkage
  int pid; // for unfreeze
  int nr_process;
  int fd, cur_fd;  // fast search
  unsigned int copied_seq;
  unsigned int saddr;
  unsigned short int sport;
  unsigned int daddr;
  unsigned short int dport;
};

struct file_header {
  unsigned long magic; /* must be == MOD_MAGIC */
  int nr_files;
  int nr_segments;
  int dumps; /* Does it contain shlib and executable dumps? */
  unsigned long vma_off; /*were memory dumps begin in this file*/
};

struct area_header {
  int nr_sections;
};

/*
at the end of the process area, it helps find the sections within
that area, by their offset
*/
struct section_header {
  int type;
  int data_size;
};

#ifdef __KERNEL__

struct process_header {
  unsigned long flags;
  int sigpending;
  unsigned long personality;
  int did_exec:1;
  pid_t pid;
  pid_t pgrp;
  pid_t tty_old_pgrp;
  pid_t session;
  pid_t tgid;
  int leader;
  int ngroups;
  pid_t family[5]; /* p_opptr, p_pptr, p_cptr, p_ysptr, p_osptr */
  uid_t uid, euid, suid, fsuid;
  gid_t gid, egid, sgid, fsgid;
  gid_t groups[NGROUPS];
  char comm[16];
  unsigned long sas_ss_sp;
  size_t sas_ss_size;
  unsigned int locks; /* file locks held */
};

struct vma {
  int is_dumped;
  unsigned long vm_start;
  unsigned long vm_end;
  unsigned long prot;
  unsigned long flags;
  unsigned long offset; /* file position */
  char filename[MAX_FILE_NAME];
};

struct memory {
  int mm_users,mm_count,map_count;
  unsigned long start_code, end_code, start_data, end_data;
  unsigned long start_brk, brk, start_stack;
  unsigned long arg_start, arg_end, env_start, env_end;
  unsigned long rss, total_vm, locked_vm;
  unsigned long def_flags;
};

struct saved_signals {
  sigset_t blocked;
  struct sigpending pending;
  struct signal_struct sig;
};

struct processor_context {
  struct thread_struct thread;
  struct pt_regs regs;
  int in_sys_call;
};
#endif

struct fdcache {
  int fd;
  struct inode *inode;
};

struct dup_file {
  int orig_fd;
};

struct pipe_file{
  unsigned long inode;
  int lock; /* Pipe lock yet to see*/
  unsigned blocked:1;
  unsigned long base;
  unsigned int start;
  int i_size;
  unsigned int rdwr; /* 0 - read, 1 - write */
};

struct regular_file{
  unsigned long int file_pos;
  unsigned long file;/* what need ? */
  int flags;
  int mode;
  char filename[MAX_FILE_NAME];
};

struct socket_file{
  int domain;
  int type;
  int protocol; /* what need ?*/
  unsigned char state;
  unsigned char reuse;
  int saddr, daddr;
  int sport, dport;
  unsigned short backlog;
  unsigned int copied_seq, snd_nxt, rcv_nxt, snd_una;
};

struct open_file{
  int entry_size; /* How many bytes are we dumping after this struct: to be changed!*/
  int type;
  int fd;
  union
  {
    struct dup_file dup;
    struct pipe_file pipes;
    struct regular_file file;
    struct socket_file sock;
  } u;
};

static inline int str_len(char* str) {
  int size = 0;
  while (str[size] != 0)
    size++;

  return size;
}


