#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <sys/time.h>

#include "../core/samory.h"
#include "../filters/filter.h"
#include "../utils/helper.h"

int main(int argc, char **argv)
{
struct command_mask mask;
int fd, pid, ret, msg_size, i, options;
char server[MAX_FILE_NAME], filename[MAX_FILE_NAME];
struct section_header section;
struct timeval time_start, time_end;


    options = 0;
    i=1;
    while (i<argc){
      if (strcmp(argv[i], "-x")==0) options |= EXEC;
      else if (strcmp(argv[i], "-l")==0) options |= SHLIB;
      else if (strcmp(argv[i], "-lx")==0) options |= (EXEC|SHLIB);
      else if (strcmp(argv[i], "-xl")==0) options |= (SHLIB|EXEC);
      else if (strncmp(argv[i], "-f", 2) == 0){
        strcpy(filename, argv[i+1]);
	options |= LCFILE;
	i+=1;
      }
      else if (strncmp(argv[i], "-s", 2) == 0){
        strcpy(server, argv[i+1]);
	options |= RRSTRT;
	i+=1;
      }
      else if(i==argc-1) pid = atoi(argv[i]);
      else {
        printf("Error in command line, please check your code\n");
        return 1;
      }
      i++;
    }

    if (options & LCFILE){
      if((fd = open(filename, O_RDWR|O_CREAT, S_IRWXU))==-1){
         perror("Can't open input output file\n");
	 return 1;
      }
    }
    else{
      if((fd = open("ckpt.file", O_RDWR|O_CREAT, S_IRWXU))==-1){
         perror("Can't open input output file\n");
	 return 1;
      }
    }

    mask.pid = pid;
    mask.fd = fd;
    mask.options = options;
    printf("checpoint: process n %d, %d\n", mask.pid, mask.options);

    if (gettimeofday(&time_start, NULL) == -1){
      fprintf(stderr, "gettimeofday error\n");
      return 1;
    }

    ret = checkpoint(&mask);
    if (ret == -1){
      printf(" Unable to open device '%s'\n", DEVICE_NAME );
      return 1;
    }
    else if (ret == -2){
      printf(" Ioctl error\n");
      return 1;
    }

    if (options & RRSTRT){
      if (transfer(fd, server) == -1){
        printf("Error in sending the file \n");
        close(fd);
        return 1;
      }
    }

    if (gettimeofday(&time_end, NULL) == -1){
      fprintf(stderr, "gettimeofday error\n");
      return 1;
    }

    printf("Checkpoint done in %d microseconds\n",get_time_elapsed(&time_start, &time_end));
    return 0;
}
