#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <wait.h>
#include <signal.h>
#include <pthread.h>

#include "../utils/samory.h"
#include "../comms/comm.h"
#include "../utils/helper.h"

// flags bits
#define	DEF_ROOT	1 	// Initializing process
#define ONE_IMG		2 	// All processes in one image
#define INCL_LIB	4 	// inludes libs in checkpoints
#define SON_ONLY	8 	// monitor the direct sons of a given process
#define ALL_FMLY	16	// monitor all descendants
#define INSTALL		32	// install also on this machine

#define SYSTEM_ROOT	"/usr/bin/samory"



/*
Gobal variables used all over the project
*/

int 			pid_list[MAX_PROCESS];//list of process i'm attached to
int 			nr_process, nr_files;	//
struct conntrack_struct cntab;
struct msg_buff		msg_buff;
struct hostent 		*host;
struct sockaddr_in 	my_addr;
struct open_file 	*open_files;
struct conns		*sks;
char 			filename[MAX_FILE_NAME];
char 			comm_file[MAX_FILE_NAME];

extern int errno;

void *install_watcher(){

int skfd, paramsfd;
struct sockaddr_in me, partner;
struct samory_command command;
struct file_header file_header;
unsigned long fcache_size;

	printf("In watching process\n");

	skfd = socket(AF_INET, SOCK_DGRAM, 0);
	paramsfd = socket(AF_INET, SOCK_DGRAM, 0);

	me.sin_family = AF_INET;
	memcpy(&me.sin_addr, host->h_addr, host->h_length);
	me.sin_port = htons(FILTER_PORT);
	bind(skfd, (struct sockaddr *)&me, sizeof(me));

	me.sin_port = htons(NEW_PARAMS_PORT);
	bind(paramsfd, (struct sockaddr *)&me, sizeof(me));

	printf("Communication sockets created\n");

	while(1){
		int r, len;
		memset(&command, 0, sizeof(command));
		printf("Entering in loop\n");
		r = recvfrom(skfd, &command, sizeof(command), 0,
			     (struct sockaddr *) &partner, &len );
		if (r < 0){
			printf("%s\n",strerror(errno));
			return;
		}
		else printf("Received %d bytes\n", r);
                switch (command.type) {
                        case ADD:{
                                pid_list[nr_process] = command.pid;
				nr_process++;
				printf("process %d added, now we have %d\n", pid_list[nr_process -1], nr_process);
                                break;
			}
                        case REM:{ //TODO see later
				printf("Not yet implmented\n");
                                break;
			}
                        case QUIT:{
				printf("All is finished, exiting...\n");
				// we null host to indicate to the filter that the system is going down
				host->h_length = 0;
				pthread_exit(NULL);
			}
                        case CKPT:{
				printf("Checkpoint process is starting...\n");
				struct command_mask mask;
				int fd, comfd, ret, i, options;
				for (i = 0; i < nr_process; i++){
					char string1[256];
					memset(string1, 0, 256);
					sprintf(string1, "%s.%d", command.filename, pid_list[i]);
					options = 0;
					if((fd = open(string1, O_RDWR|O_CREAT, S_IRWXU))==-1){
        					printf("%s\n",strerror(errno));
						return ;
					}
					mask.pid = pid_list[i];
					mask.fd = fd;
					mask.options = (EXEC|SHLIB);
					printf("checkpointing pid n %d\n", mask.pid);
					ret = checkpoint(&mask);
					if (ret == -1){
						printf(" Unable to open device '%s'\n", DEVICE_NAME );
						return ;
					}
					else if (ret == -2){
						printf(" Ioctl error\n");
						return;
					}
					close(fd);
					// checkpoint done, now go to the messages

					fd = open(string1, O_RDONLY);

					if( read(fd, &file_header,sizeof(file_header))!=sizeof(file_header)){
						printf("Did not receive enough data\n");
						return;
					}
					if (file_header.magic != MOD_MAGIC) {
						printf("Not a recognized file\n");
						return;
					}

					fcache_size = file_header.nr_files*sizeof(struct open_file);
					open_files = (struct open_file *)malloc(fcache_size);
					if (open_files == NULL){
						printf("Cannot allocate cache space\n");
						close(fd);
						return;
					}

					if( read(fd, open_files,fcache_size)!=fcache_size){
						printf("Did not receive enough data for open files\n");
						close(fd);
						return;
					}

					if((comfd = open(command.comm_file, O_RDWR|O_CREAT, S_IRWXU))==-1){
        					printf("%s\n",strerror(errno));
						return;
					}
					store_coms(&cntab, &msg_buff, open_files, file_header.nr_files, comfd, skfd);
					close(fd);
					close(comfd);
					free(open_files);
				}
				break;

			}
                        case RST:{
				printf("Restarting a process image file is %s, and comm file is %s\n",
				command.filename, command.comm_file);
				struct restart_param param;
				int fd, comfd, ret, i, pid, options, nr_sockets;

				if((fd = open(command.filename, O_RDONLY))==-1){
				       perror("Can't open restart image file\n");
				       return;
				}

				if( read(fd, &file_header,sizeof(file_header))!=sizeof(file_header)){
					printf("Did not receive enough data\n");
					return;
				}
				if (file_header.magic != MOD_MAGIC) {
					printf("Bad file format \n");
					return;
				}
				printf("%s, good format, %d open files, %d segments\n", command.filename,
					file_header.nr_files, file_header.nr_segments);

				fcache_size = file_header.nr_files*sizeof(struct open_file);
				open_files = (struct open_file *)malloc(fcache_size);

				if (open_files == NULL){
					printf("Cannot allocate cache space for open files\n");
					close(fd);
					return;
				}

				if( read(fd, open_files,fcache_size)!=fcache_size){
					printf("Files data do not match open files number\n");
					close(fd);
					return;
				}

				close(fd);
				printf("Opening communication file\n");
				if((comfd = open(command.comm_file, O_RDONLY))==-1){
					perror("Can't open communication input file\n");
					return;
				}
				printf("Restoring communications\n");

				// rebuild the connection table and message queues
				ret = rebuild_conntable(comfd, &cntab, &msg_buff);
				close(comfd);

				if(ret == 0){
					printf("Error when rebuilding connection table\n");
					return;
				}

				printf("Updating sockets params\n");
				sks = (struct conns *)malloc(sizeof(struct conns) * MAX_SOCKETS);
				if(sks == NULL){
					printf("Not enough space for socket table\n");
					return;
				}
				bzero(sks, sizeof(struct conns)*MAX_SOCKETS);
				// Apply kernel params to the connections
				ret = verify_kernel_params(open_files, file_header.nr_files, &cntab, sks);
				if(ret == -1){
					printf("Error in connection table restablished\n");
					return;
				}
				printf("Done, continue\n");
				// Tell the remotes that migration is beginning
				printf("Restoring the sockets\n");
				nr_sockets = restore_sockets(skfd, paramsfd, sks, &cntab, ret);
				if(nr_sockets != ret){
					printf("Error in restoring sockets\n");
					return;
				}
				printf("Sockets restored\n");
				pid = getpid();
				for(i = 0; i<nr_sockets; i++)
					sks[i].pid = pid;

				if ((pid = fork())<0){
					printf("Fork error, exiting\n");
					return;
				}

				if (pid == 0){

//					char system_command[256];

				 // this process takes on the restart procedure
				// Arriving here, all the conntected sockets have been restarted
				// sks contains the newly created sockets and link to corresponding connection
				// each process has to look in that table and steal a socket, corresponding to its state
					close(skfd);
					close(paramsfd);

/*					for(i = 0; i<nr_sockets; i++)
					close(sks[i].fd);*/

					// at this stage we have closed all unnecessary files
					// all necessary ones (established sockets) are in sks

					reopen_files(open_files, file_header.nr_files, nr_sockets, sks, host);
					replay_msg(sks, &cntab, &msg_buff, nr_sockets);
					printf("Files restored, continuing...\n");

/*					memset(system_command, 0, 256);
					strcpy(system_command, "xterm -e ./restart ");
					strcat(system_command, command.filename);
//					system(system_command);
					printf("Type ./restart %s to restart your process in a shell\n", command.filename);*/

					param.options = RESTART;
					param.pid = getppid();
					strcpy(param.ckpt_file, command.filename);

    					if ((restart(&param) < 0)){
	   					printf("An error occured when restarting the process\n");
	   					return;
					}
					return ;
				}
				else{
					pid_list[nr_process] = pid;
					nr_process++;
					printf("process %d added, now we have %d\n",pid_list[nr_process -1], nr_process);
				// unlink the buffers concerned by this process and free that space.
					zap_buffers(sks, &cntab, &msg_buff, nr_sockets);
					free(open_files);
					free(sks);
				}
				break;

			}
                        case MIGR:{
				printf("About to migrate the system\n");
				struct hostent *d_host;
				d_host = gethostbyname(command.dest_host);
				partner.sin_family = AF_INET;
				memcpy(&partner.sin_addr, d_host->h_addr, d_host->h_length);
				partner.sin_port = htons(FILTER_PORT);
				command.type = RST;

				sendto(skfd, &command, sizeof(command), 0,
					(struct sockaddr *)&partner, sizeof(partner));
				break;
			}
                        case RST_MSG :{
				printf("Other has migrated, updating our tables\n");
				handle_restart_notify(paramsfd, &cntab, &msg_buff, nr_process, pid_list);
				break;
			}
			case CHK_MSG :{
				printf("Connection update\n");
				handle_notify(0, (struct msg_struct *)&command, &cntab, NULL, 0, NULL);
				break;
			}
                        default :{
				printf("Unknown command\n");
				break;
			}

		}
	}
	pthread_exit(0);
}


void init(char *config_file, int flags){
FILE *fd;
pthread_t thr;

struct r_command{
	char path[26];
	char address[16];
	char option[5];
}r_;

int ret, pid;
extern int errno;
char hname[256];

	if (flags & DEF_ROOT){
		printf("In dispatcher\n");
		fd = fopen(config_file, "r");
		memset(&r_,0,sizeof(r_));
		memcpy(&r_.path, SYSTEM_ROOT,26);
		memcpy(&r_.option, "init", 4);
		while (fscanf(fd, "%s", &r_.address) != EOF){
			pid = fork();
			if (pid == 0){

			}
			else if (pid > 0){
				wait(NULL);
				memset(&r_.address, 0, 16);
			}
			else
				break;
		}
		fclose(fd);
	}

	if((flags & INSTALL) == 0){
		printf("Not to be installed on this machine, exiting\n");
		return;
	}

	printf("Initializing the global parameters\n");

	memset(pid_list, 0, sizeof(int)*MAX_PROCESS);//list of process i'm attached to
	nr_process = nr_files = 0;
	memset(&cntab, 0, sizeof(cntab));	// Table of connections
	cntab.next = NULL;
	memset(&msg_buff, 0, sizeof(msg_buff));

	if(gethostname(hname, 256) == 0){
		printf("found host %s\n", hname);
		host = gethostbyname(hname);
	}
	else return;

	if(host==NULL){
		printf("Null host, exiting\n");
		return;
	}

	open_files = NULL;
//	memset(sks, 0, sizeof(struct conns)*MAX_SOCKETS);
	memset(filename, 0, MAX_FILE_NAME);
	memset(comm_file, 0, MAX_FILE_NAME);

	printf("Done, Now spawn a listening thread for communication\n");
	ret = pthread_create(&thr, NULL, install_watcher, NULL);
	if(ret != 0){
		printf("Error in thread creation\n");
		printf("%s\n",strerror(ret));
		kill(pid, SIGKILL);
		return;
	}

	printf(" Loading message filter\n");

	install_filter(host, &cntab, &msg_buff, pid_list, &nr_process);

	// if this function returns, host is NULL that means we have to quit
	pthread_join(thr, NULL);
	return; // if you end someday ?
}


int main(int argc, char **argv){
char config_file[256];
int flags, tmpfd, i;
char hname[256];

struct samory_command command;

	flags = 0;
	if (strcmp(argv[1], "init") == 0){
		i = 2;
		printf("Loking for a running instance\n");
		if((tmpfd = open("/tmp/.samory.presence", O_EXCL|O_CREAT)) <= 0){
			printf("Already running: first exit or modify parmeters\n");
			printf("%s\n",strerror(errno));
			return 0;
		}
		printf("Not running, Entering initialisation process\n");
		memset(config_file, 0, 256);
		strcpy(config_file, "samory.machines");
		printf("treating args\n");
		while (i < argc){
			if (strcmp(argv[i], "-c") == 0){
				memset(config_file, 0, 256);
				strcpy(config_file, argv[i+1]);
				i+=1;
			}
			else if (strcmp(argv[i], "-i") == 0) flags |= INSTALL;
			else if (strcmp(argv[i], "-r") == 0) flags |= DEF_ROOT;
			else if ((strcmp(argv[i], "-ir") == 0) ||
				 (strcmp(argv[i], "-ri") == 0)) flags |= (INSTALL | DEF_ROOT);

			i++;
		}
		printf("calling init func\n");
		close(tmpfd);
		init(config_file, flags);
		printf("System stopped, exiting\n");
		remove("/tmp/.samory.presence");
	}

	else{
	int skfd, size;
	struct sockaddr_in me, instance;
	// When an instance is running this only serves as a feeder for that instance
		if((tmpfd = open("/tmp/.samory.presence", O_EXCL|O_CREAT)) >= 0){
			printf("You have to init first\n");
			remove("/tmp/.samory.presence");
			return 0;
		}

		printf("Running\n");
		if (strcmp(argv[1], "add")==0){
			printf("ADD command detected\n");
			command.type = ADD;
			command.pid = atoi(argv[2]);
		}

		if (strcmp(argv[1], "rem")==0){
			command.type = REM;
			command.pid = atoi(argv[2]);
		}

		if (strcmp(argv[1], "quit")==0){
			command.type = QUIT;
		}

		if (strcmp(argv[1], "checkpoint")==0){
			command.type = CKPT;
			memset(command.filename, 0, MAX_FILE_NAME);
			memset(command.comm_file, 0, MAX_FILE_NAME);
			memcpy(command.filename, argv[2], strlen(argv[2]));
			memcpy(command.comm_file, argv[3], strlen(argv[3]));
		}

		if (strcmp(argv[1], "restart")==0){
			command.type = RST;
//			command.pid = 0;
			memset(command.filename, 0, MAX_FILE_NAME);
			memset(command.comm_file, 0, MAX_FILE_NAME);
//			memset(command.dest_host, 0, 16);
			memcpy(command.filename, argv[2], strlen(argv[2]));
			memcpy(command.comm_file, argv[3], strlen(argv[3]));
//			memcpy(command.dest_host, ttyname(0), strlen(ttyname(0)));
		}

		if (strcmp(argv[1], "migrate")==0){
			command.type = MIGR;
			memset(command.filename, 0, MAX_FILE_NAME);
			memset(command.comm_file, 0, MAX_FILE_NAME);
			memset(command.dest_host, 0, 16);
			memcpy(command.filename, argv[2], strlen(argv[2]));
			memcpy(command.comm_file, argv[3], strlen(argv[3]));
			memcpy(command.dest_host, argv[4], strlen(argv[4]));
		}

	// Now bind a UDP socket to the communication channel,
		skfd = socket(AF_INET, SOCK_DGRAM, 0);
		me.sin_family = AF_INET;
		me.sin_addr.s_addr = htonl(INADDR_ANY);
		me.sin_port = htons(0);

		bind(skfd, (struct sockaddr *)&me, sizeof(me));

		if(gethostname(hname, 256) == 0){
			printf("found host %s\n", hname);
			host = gethostbyname(hname);
		}
		else return;

		if(host==NULL){
			printf("Null host, exiting\n");
			return;
		}

		instance.sin_family = AF_INET;
		memcpy(&instance.sin_addr, host->h_addr, host->h_length);
		instance.sin_port = htons(FILTER_PORT);

		size = sendto(skfd, &command, sizeof(command), 0,
		         (struct sockaddr *)&instance, sizeof(instance));
		if(size < sizeof(command)){
			fprintf(stderr, "send error\n");
			printf("%s\n",strerror(errno));
			return 0;
		}
	}
	return 0;
}

